# coding=utf-8
"""A jumble of seemingly useful stuff."""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
from functools import partial
from itertools import chain
import datetime
import inspect
from collections import OrderedDict
from socket import gethostname


# http://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits
MAX_EXT4_FN_LENGTH = 255


# --- Introspection tools

def curry2partial(obj):
    """Transforms toolz/cytoolz curry sugar into standard partials."""
    try:
        return partial(obj.func, *obj.args, **obj.keywords)
    except AttributeError:
        return obj


def callable2call(c, closure_extractor=lambda c: c):
    """
    Extracts the (actual) function name and set parameters from a callable.

    Manages partials and makes a best effort to manage anonymous functions and non-introspectable objects.


    Parameters
    ----------
    c : callable
        The function, partial, builtin or, in general, any callable.

    closure_extractor : function c -> function, default c -> c
        A function that will be executed whenever c is a closure.
        It should return a function to be analyzed by inspect.
        This can be useful to specify custom behavior with, e.g., decorated functions

    Returns
    -------
    a pair (function_name, parameter_dict), where the parameter_dict is a
    (possibly empty) dictionary {parameter_name: default_value} with the parameter
    values that are known by either default values or the partial application of the function.

    *N.B. these parameters are fixed iff they are set by a partial application, otherwise,
    for default parameters, it is up to the caller not to redefine them.*

    Examples
    --------
    >>> callable2call(partial)
    ('partial', {})

    >>> callable2call(partial(map))
    ('map', {})

    >>> name, params = callable2call(partial(map, function=str))
    >>> name
    'map'
    >>> 'str' in str(params['function'])
    True

    >>> name, params = callable2call(partial(partial(map, function=str), iterable1=()))
    >>> name
    'map'
    >>> 'str' in str(params['function'])
    True
    >>> () == params['iterable1']
    True

    >>> name, params = callable2call(lambda x: x)
    >>> print(name)
    lambda
    >>> params
    {}

    >>> name, params = callable2call(lambda x=5: x)
    >>> print(name)
    lambda
    >>> params
    {'x': 5}
    """
    def callable2call_recursive(c, positional=None, keywords=None):
        if keywords is None:
            keywords = {}
        if positional is None:
            positional = []
        if inspect.isfunction(c):
            if is_closure(c):  # allow custom behavior with closures
                c = closure_extractor(c)
            args, _, _, defaults = inspect.getargspec(c)
            defaults = [] if not defaults else defaults
            args = [] if not args else args
            args_set = set(args)
            # Check that everything is fine...
            keywords = dict(chain(zip(args[-len(defaults):], defaults), keywords.items()))  # N.B. order matters
            pos2keyword = dict(zip(args[:len(positional)], positional))  # N.B. order matters
            keywords_set = set(keywords.keys())
            if len(keywords_set - args_set) > 0:
                raise ValueError('Some partial %r keywords are not parameters of the function %s' %
                                 (keywords_set - args_set, c.__name__))
            if len(set(pos2keyword.keys()) & set(keywords.keys())) > 0:
                raise ValueError('Some arguments are indicated both by position and name (%r)' %
                                 sorted(set(pos2keyword.keys()) & set(keywords.keys())))
            if len(args_set) - len(keywords_set) < len(positional):
                raise ValueError('There are too many positional arguments indicated '
                                 'for the number of unbound positional parameters left.')
            return c.__name__.replace('<', '').replace('>', ''), \
                dict(list(keywords.items()) + list(pos2keyword.items()))
        if isinstance(c, partial):
            pkeywords = c.keywords if c.keywords is not None else {}
            return callable2call_recursive(
                c.func,
                positional=positional + list(c.args),                  # N.B. order matters
                keywords=dict(chain(pkeywords.items(), keywords.items())))   # N.B. order matters
        if hasattr(c, '__call__'):
            # No way to get the argspec from anything arriving here (builtins and the like...)
            return c.__name__, keywords
        raise ValueError('Only callables (partials, functions, builtins...) are allowed, %r is none of them' % c)
    return callable2call_recursive(c)


def all_subclasses(cls):
    """Returns a list with all the subclasses of cls in the current python session.

    Examples
    --------
    Some C inheritance...
    >>> bool in all_subclasses(int)
    True
    >>> int in all_subclasses(bool)
    False

    This is recursive...
    >>> class my_str(str):
    ...     pass
    >>> my_str in all_subclasses(str)
    True
    """
    return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in all_subclasses(s)]


def _dict(obj):
    """Returns a copy of obj.__dict___ (or {} if obj has not __dict__).

    Examples
    --------
    >>> from future.moves.collections import UserDict
    >>> _dict(UserDict())
    {'data': {}}
    >>> class NoSlots(object):
    ...     def __init__(self):
    ...         self.prop = 3
    >>> ns = NoSlots()
    >>> _dict(ns)
    {'prop': 3}
    >>> _dict(ns) is not ns.__dict__
    True
    >>> class Slots(object):
    ...     __slots__ = ['prop']
    ...     def __init__(self):
    ...         self.prop = 3
    >>> _dict(Slots())
    {}
    """
    try:
        return obj.__dict__.copy()
    except:
        return {}


def _slotsdict(obj):
    """Returns a dictionary with all attributes in obj.__slots___ (or {} if obj has not __slots__).

    Examples
    --------
    >>> from future.moves.collections import UserDict
    >>> _slotsdict(UserDict())
    {}
    >>> class Slots(object):
    ...     __slots__ = 'prop'
    ...     def __init__(self):
    ...         self.prop = 3
    ...     @property
    ...     def prop2(self):
    ...         return 5
    >>> _slotsdict(Slots())
    {'prop': 3}
    """
    descriptors = inspect.getmembers(obj.__class__, inspect.isdatadescriptor)
    return {dname: value.__get__(obj) for dname, value in descriptors if
            '__weakref__' != dname if inspect.ismemberdescriptor(value)}


def _propsdict(obj):
    """Returns the @properties in an object.
    Example:
    >>> class PropertyCarrier(object):
    ...     __slots__ = 'prop2'
    ...     def __init__(self):
    ...         self.prop2 = 3
    ...     @property
    ...     def prop(self):
    ...         return self.prop2
    ...     @prop.setter
    ...     def prop(self, prop):
    ...         self.prop2 = prop
    >>> _propsdict(PropertyCarrier())
    {'prop': 3}
    """
    descriptors = inspect.getmembers(obj.__class__, inspect.isdatadescriptor)
    # All data descriptors except slots and __weakref__
    # See: http://docs.python.org/2/reference/datamodel.html
    return {dname: value.__get__(obj) for dname, value in descriptors if
            '__weakref__' != dname and not inspect.ismemberdescriptor(value)}


def trim_dict(cd, exclude_prefix='_', exclude_postfix='_', excludes=('what',)):
    """Removes keys from a dictionary if they are (pre/post)fixed or fully match forbidden strings.

    Parameters
    ----------
    cd: dictionary
        The dictionary to trim

    exclude_prefix: string, default '_'
        Exclude all attributes whose name starts with this string

    exclude_postix: string, default '_'
        Exclude all attributes whose name ends with this string

    excludes: string iterable, default ('what',)
        Exclude all attributes whose name appears in this collection

    Returns
    -------
    A copy of cd with only the allowed keys.
    """
    return {k: v for k, v in cd.items() if
            (exclude_prefix and not k.startswith(exclude_prefix)) and
            (exclude_postfix and not k.endswith(exclude_postfix)) and
            k not in set(excludes)}


def config_dict_for_object(obj,
                           add_dict=True,
                           add_slots=True,
                           add_properties=True):
    """Returns a dictionary with obj attributes defined in __dict__, __slots__ or as @properties.
    Does not fail in case any of these are not defined.

    Parameters
    ----------
    obj: anything
        The object to introspect

    add_dict: boolean, default True
        Add all the attributes defined in obj.__dict__

    add_slots: boolean, default True
        Add all the attributes defined in obj.__slots__

    add_properties: boolean, default True
        Add all the attributes defined as obj @properties

    Returns
    -------
    A dictionary {atribute: value}

    Examples
    --------
    >>> class NoSlots(object):
    ...     def __init__(self):
    ...         self.prop = 3
    ...         self._hidden = 5
    ...         self.hidden_ = 5
    >>> class Slots(NoSlots):
    ...     __slots__ = 'sprop'
    ...     def __init__(self):
    ...         super(Slots, self).__init__()
    ...         self.sprop = 4
    >>> class Props(Slots):
    ...     def __init__(self):
    ...         super(Props, self).__init__()
    ...     @property
    ...     def pprop(self):
    ...         return 5
    >>> obj = Props()
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=False, add_slots=False, add_properties=False)).items())
    []
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=True, add_slots=False, add_properties=False)).items())
    [('prop', 3)]
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=True, add_slots=True, add_properties=False)).items())
    [('prop', 3), ('sprop', 4)]
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=True, add_slots=False, add_properties=True)).items())
    [('pprop', 5), ('prop', 3)]
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=True, add_slots=True, add_properties=True)).items())
    [('pprop', 5), ('prop', 3), ('sprop', 4)]
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=False, add_slots=True, add_properties=False)).items())
    [('sprop', 4)]
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=False, add_slots=False, add_properties=True)).items())
    [('pprop', 5)]
    >>> sorted(trim_dict(config_dict_for_object(obj, add_dict=False, add_slots=True, add_properties=True)).items())
    [('pprop', 5), ('sprop', 4)]
    """
    # see also dir
    cd = {}
    if add_dict:
        cd.update(_dict(obj))
    if add_slots:
        cd.update(_slotsdict(obj))
    if add_properties:
        cd.update(_propsdict(obj))
    return cd


def is_closure(c):
    """Checks whether an object is a python closure."""
    return inspect.isfunction(c) and c.__closure__ is not None


def extract_decorated_function_from_closure(c):
    """
    Extracts a function from closure c iff the closure only have one cell mapping to a function
    and also has a method 'what'.
    (this ad-hoc behavior could help to play well with the whatable decorator)
    """
    closure = c.__closure__
    if closure is not None and len(closure) == 1 and hasattr(c, 'what'):
        func = closure[0].cell_contents
        if inspect.isfunction(func):
            return extract_decorated_function_from_closure(func)
    return c


def is_iterable(v):
    """Checks whether an object is iterable or not."""
    try:
        iter(v)
    except:
        return False
    return True


# --- Tools to make aggregation of information from functions and whatables easier

def mlexp_info_helper(title,
                      data_setup=None,
                      model_setup=None,
                      eval_setup=None,
                      exp_function=None,
                      comments=None,
                      itime=None):
    """Creates a dictionary describing machine learning experiments.

    Parameters:
      - title: the title for the experiment
      - data_setup: a "what" for the data used in the experiment
      - model_setup: a "what" for the model used in the experiment
      - eval_setup: a "what" for the evaluation method used in the experiment
      - exp_function: the function in which the experiment is defined;
                      its source text lines will be stored
      - comments: a string with whatever else we need to say
      - itime: a callable that will provide the date from a reliable internet source
               (e.g. an NTP server or http://tycho.usno.navy.mil/cgi-bin/timer.pl)

    (Here "what" means None, a string, an object providing a "what method" or an object providing an "id" method.)

    Return:
      An ordered dict mapping strings to strings with all or part of:
      title, data_setup, model_setup, eval_setup, fsource, date, idate (internet datetime), host, comments
    """
    from .whatutils import what2id
    info = OrderedDict((
        ('title', title),
        ('data_setup', what2id(data_setup)),
        ('model_setup', what2id(model_setup)),
        ('eval_setup', what2id(eval_setup)),
        ('fsource', inspect.getsourcelines(exp_function) if exp_function else None),
        ('date', datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')),
        ('idate', None if itime is None else itime()),
        ('host', gethostname()),
        ('comments', comments),
    ))
    return info



# coding=utf-8
from functools import partial

import pickle

from ..what import is_whatable, What, trim_dict, config_dict_for_object
from .fixtures import *


def test_whatable_decorator():
    @whatable
    def normalize(x, loc=5, scale=3):
        """returns (x+loc) / scale"""
        return (x + loc) / scale
    assert normalize.what().id() == 'normalize(loc=5,scale=3)'
    assert normalize.__name__ == 'normalize'
    assert normalize.__doc__ == 'returns (x+loc) / scale'

    # very specific case: partial application over a whatable closure
    normalize6 = partial(normalize, loc=6)
    assert not hasattr(normalize6, 'what')
    assert not hasattr(normalize6, '__name__')  # partials have no name
    normalize6 = whatable(normalize6)
    assert normalize6.what().id() == 'normalize(loc=6,scale=3)'
    assert normalize6(3) == 3


def test_whatable_simple(c1):
    # Non-nested configurations
    config_c1 = c1.what()
    assert config_c1.name == 'C1'
    assert len(config_c1.conf) == 3
    assert config_c1.conf['p1'] == 'blah'
    assert config_c1.conf['p2'] == 'bleh'
    assert config_c1.conf['length'] == 1
    assert config_c1 == config_c1
    assert config_c1.id() == "C1(length=1,p1='blah',p2='bleh')"


def test_nested_whatables(c1, c2):
    # Nested whatables
    config_c2 = c2.what()
    assert config_c2.name == 'C2'
    assert len(config_c2.conf) == 2
    assert config_c2.conf['name'] == 'roxanne'
    assert config_c2.conf['c1'].what() == c1.what()
    assert config_c2.id() == "C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne')"
    c2.c1 = c1.what()
    config_c2 = c2.what()
    assert config_c2.id() == "C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne')"


def test_whatable_partial(c1):

    def identity(x):
        return x

    # Partial functions
    c1.p1 = partial(identity, x=1)
    assert c1.what().id() == "C1(length=1,p1=identity(x=1),p2='bleh')"
    assert c1.p1() == 1


def test_whatable_builtins(c1):
    # Builtins - or whatever foreigner - do not allow introspection
    c1.p1 = sorted
    with pytest.raises(Exception) as excinfo:
        c1.what().id()
    assert str(excinfo.value) == 'Cannot determine the argspec of a non-python function (sorted). ' \
                                 'Please wrap it in a whatable'


def test_whatable_anyobject(c1):

    # Objects without proper representation
    class RandomClass(object):
        def __init__(self):
            self.param = 'yes'
    c1.p1 = RandomClass()
    assert c1.what().id() == "C1(length=1,p1=RandomClass(param='yes'),p2='bleh')"


def test_whatable_force_flag():
    @whatable(force_flag_as_whatami=True)
    class A(object):

        def __init__(self):
            super(A, self).__init__()
            self.a = 3

        def what(self):
            return whatareyou(self)

    assert is_whatable(A)
    assert is_whatable(A())
    assert A().what().id() == 'A(a=3)'


def test_whatable_data_descriptors():

    # Objects with @properties
    class ClassWithProps(object):
        def __init__(self):
            self._prop = 3

        @property
        def prop(self):
            return self._prop

        def what(self):
            return What(self.__class__.__name__,
                        trim_dict(config_dict_for_object(self, add_properties=True)))

    cp = ClassWithProps()
    assert cp.what().id() == 'ClassWithProps(prop=3)'

    # Objects with dynamically added properties
    setattr(cp, 'dprop', property(lambda: 5))
    with pytest.raises(Exception) as excinfo:
        cp.what().id()
    assert str(excinfo.value) == 'Dynamic properties are not suppported.'


def test_is_whatable(c1):
    assert is_whatable(c1)
    assert not is_whatable(str)


def test_whatable_custom_func():

    def whatfunc(obj):  # pragma: no cover
        return What('custom', conf={'n_trees': None, 'original': obj.__name__})

    def rfc(n_trees=30):  # pragma: no cover
        return n_trees

    assert whatable(rfc).what().id() == 'rfc(n_trees=30)'
    assert whatable(rfc, whatfunc=whatfunc).what().id() == 'custom(n_trees=None,original=\'rfc\')'


@pytest.mark.xfail(reason='fix infinite recursion')
def test_whatable_custom_func_recursive():

    def whatfunc(obj):  # pragma: no cover
        return What('custom', conf={'n_trees': None, 'original': obj})

    def rfc(n_trees=30):  # pragma: no cover
        return n_trees

    assert whatable(rfc).what().id() == 'rfc(n_trees=30)'
    assert whatable(rfc, whatfunc=whatfunc).what().id() == 'custom(n_trees=None,original=rfc(n_trees=30))'


def test_whatable_slots():
    # N.B. Slots are implemented as descriptors
    @whatable
    class Slots(object):
        __slots__ = ['prop']

        def __init__(self):
            self.prop = 3

    slots = Slots()
    assert slots.what().id() == 'Slots(prop=3)'


def test_whatable_inheritance():
    # Inheritance works as spected
    @whatable
    class Super(object):
        def __init__(self):
            super(Super, self).__init__()
            self.a = 'superA'
            self.b = 'superB'

    class Sub(Super):
        def __init__(self):
            super(Sub, self).__init__()
            self.c = 'subC'
            self.a = 'subA'

    assert Sub().what().id() == "Sub(a='subA',b='superB',c='subC')"


def test_whatable_does_not_override_what(c1):
    c1.what = 33
    assert not is_whatable(c1)
    with pytest.raises(Exception) as excinfo:
        whatable(c1)
    assert str(excinfo.value) == 'object already has an attribute what, and is not a whatami what, ' \
                                 'if you know what I mean'


def test_whatable_torturing_inheritance():

    class D1(object):

        def __init__(self):
            self.d1 = 1

    class S1(D1):
        __slots__ = 's1'

        def __init__(self):
            super(S1, self).__init__()
            self.s1 = 2

    class S2(S1):
        __slots__ = 's2'

        def __init__(self):
            super(S2, self).__init__()
            self.s2 = 3

    class D2(S2):

        def __init__(self):
            super(D2, self).__init__()
            self.d2 = 4

    class P1(D2):

        def __init__(self):
            super(P1, self).__init__()

        @property
        def p1(self):
            return 5

    class S3(P1):
        __slots__ = 's3'

        def __init__(self):
            super(S3, self).__init__()
            self.s3 = 6

    s3 = S3()

    s3 = whatable(s3, add_dict=True, add_slots=True, add_properties=True)
    assert s3.what().id() == "S3(d1=1,d2=4,p1=5,s1=2,s2=3,s3=6)"

    s3 = whatable(s3, add_dict=True, add_slots=True, add_properties=False)
    assert s3.what().id() == "S3(d1=1,d2=4,s1=2,s2=3,s3=6)"

    s3 = whatable(s3, add_dict=True, add_slots=False, add_properties=True)
    assert s3.what().id() == "S3(d1=1,d2=4,p1=5)"

    s3 = whatable(s3, add_dict=True, add_slots=False, add_properties=False)
    assert s3.what().id() == "S3(d1=1,d2=4)"

    s3 = whatable(s3, add_dict=False, add_slots=False, add_properties=False)
    assert s3.what().id() == "S3()"

    s3 = whatable(s3, add_dict=False, add_slots=True, add_properties=False)
    assert s3.what().id() == "S3(s1=2,s2=3,s3=6)"

    s3 = whatable(s3, add_dict=False, add_slots=False, add_properties=True)
    assert s3.what().id() == "S3(p1=5)"

    s3 = whatable(s3, add_dict=False, add_slots=True, add_properties=True)
    assert s3.what().id() == "S3(p1=5,s1=2,s2=3,s3=6)"


def test_whatable_duck():

    class DuckedWhatable(object):
        def what(self):
            return What(self.__class__.__name__, {'param1': 33})
    cduck = DuckedWhatable()
    assert cduck.what().id() == 'DuckedWhatable(param1=33)'

    @whatable
    class NestedDuckedWhatable(object):
        def __init__(self):
            super(NestedDuckedWhatable, self).__init__()
            self.ducked = cduck
    nested_duck = NestedDuckedWhatable()
    assert nested_duck.what().id() == 'NestedDuckedWhatable(ducked=DuckedWhatable(param1=33))'


def test_whatable_builtin():
    with pytest.raises(TypeError) as excinfo:
        whatable(all)
    assert 'builtins cannot be whatamised' in str(excinfo.value)

    with pytest.raises(Exception) as excinfo:
        whatable(str)
    assert 'cannot whatamise' in str(excinfo.value)


def test_whatable_faker():

    class Faker(object):
        def what(self):
            return 'Fool you; %s' % self.__class__.__name__

    @whatable
    class Fool(object):
        def __init__(self):
            self.faker = Faker()

    assert 'Fool you; Faker' not in whatareyou(Fool()).id()
    assert whatareyou(Fool()).id() == 'Fool(faker=Faker())'


# --- Test pickability (https://github.com/sdvillal/whatami/issues/8)

def pickle_roundtrip(x):
    return pickle.loads(pickle.dumps(x))


class Pickable(object):
    def __init__(self, x=3):
        super(Pickable, self).__init__()
        self.x = x


def test_whatable_class_pickling():

    # Decorating a Pickable class
    WhatablePickable = pickle_roundtrip(whatable(Pickable))
    assert WhatablePickable.__name__ == 'Pickable'
    assert WhatablePickable(5).what().id() == 'Pickable(x=5)'


def pickable(x, y, z=3):  # pragma: no cover
    return x + y + z


@pytest.mark.xfail(reason='need to find a way to do this')
def test_whatable_function_pickling():  # pragma: no cover
    whatable_pickable = whatable(pickable)
    assert whatable_pickable.what().id() == 'pickable(z=3)'
    roundtripped = pickle_roundtrip(whatable_pickable)
    assert roundtripped.what().id() == 'pickable(z=3)'


@pytest.mark.xfail(reason='allow to patch types and, if not, add at least a whatable partial construct')
def test_whatable_type():  # pragma: no cover
    whatable_partial = whatable(partial)
    m = whatable_partial(pickable, x=1)
    assert m.what().id() == 'pickable(x=1,z=3)'
    assert m(y=2) == 6
    whatable_partial = pickle_roundtrip(whatable_partial)
    m = whatable_partial(pickable, x=1)
    assert m.what().id() == 'pickable(x=1,z=3)'
    assert m(y=2) == 6


def test_whatable_inplacefunc():

    def afunc(x=3):  # pragma: no cover
        return x

    wafunc = whatable(afunc, modify_func_inplace=True)

    assert afunc is wafunc
    assert afunc.what().id() == wafunc.what().id()
    assert wafunc.what().id() == 'afunc(x=3)'



# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import print_function, unicode_literals, absolute_import
import inspect
from functools import partial

from future.builtins import str as str3

from .what import What, whatareyou
from .misc import callable2call, config_dict_for_object, curry2partial


# --- Basic plugins

def what_plugin(_, v):
    """Deals with What objects."""
    if isinstance(v, What):
        return v.id()
    return None


def whatable_plugin(_, v):
    """Deals with whatable objects."""
    if hasattr(v, 'what'):
        what = v.what
        what = what() if callable(what) else what
        return what_plugin(None, what)


def builtin_plugin(_, v):
    """Special message if we try to pass something like sorted or np.array"""
    if inspect.isbuiltin(v):
        raise Exception('Cannot determine the argspec of a non-python function (%s). '
                        'Please wrap it in a whatable' % v.__name__)


def property_plugin(_, v):
    """Deals with dynamic properties."""
    if isinstance(v, property):
        raise Exception('Dynamic properties are not suppported.')


def dict_plugin(what, v):
    if isinstance(v, dict):
        kvs = sorted('%s:%s' % (what.build_string(k), what.build_string(v)) for k, v in v.items())
        return '{%s}' % ','.join(kvs)
    return None


def set_plugin(what, v):
    if isinstance(v, set):
        elements = sorted(map(what.build_string, v))
        return '{%s}' % ','.join(elements) if len(elements) > 0 else 'set()'
    return None


def list_plugin(what, v):
    if isinstance(v, list):
        return '[%s]' % ','.join(map(what.build_string, v))
    return None


def tuple_plugin(what, v):
    if isinstance(v, tuple):
        return '(%s)' % ','.join(map(what.build_string, v))
    return None


def string_plugin(_, v):
    if isinstance(v, (str, str3)):
        return '\'%s\'' % v
    return None


# --- Function plugins


def partial_plugin(what, v):
    """Deals with partials and toolz curried functions.
    Configuration are the set parameters ("compulsory") and default parameters ("weak").
    """
    v = curry2partial(v)
    if isinstance(v, partial):
        name, keywords = callable2call(v)
        return What(name, keywords, what.non_id_keys).id()
    return None


def function_plugin(what, v):
    """Deals with functions, configuration are the keyword args."""
    if inspect.isfunction(v):
        args, _, _, defaults = inspect.getargspec(v)
        defaults = [] if not defaults else defaults
        args = [] if not args else args
        params_with_defaults = dict(zip(args[-len(defaults):], defaults))
        name = v.__name__ if v.__name__ != '<lambda>' else 'lambda'
        what = What(name, params_with_defaults, what.non_id_keys)
        return what.id()


# --- "Capture all" plugins

def anyobject0x_plugin(what, v):
    """An object without proper representation, try a best effort."""
    if ' at 0x' in str(v):
        what = What(v.__class__.__name__, config_dict_for_object(v), what.non_id_keys)
        return what.id()
    return None


def anyobject_plugin(_, v):
    """Delegate to str, this should be the last plugin in the chain."""
    return str(v)

# --- Numpy and pandas

try:  # pragma: no cover
    from joblib.hashing import hash as hasher
    hasher = partial(hasher, hash_name='md5')
except ImportError:  # pragma: no cover
    hasher = None


def has_joblib():
    """Returns True iff joblib can be imported."""
    return hasher is not None


try:  # pragma: no cover
    import numpy as np
except ImportError:  # pragma: no cover
    np = None


def has_numpy():
    """Returns True iff numpy can be imported."""
    return np is not None


try:  # pragma: no cover
    import pandas as pd
except ImportError:  # pragma: no cover
    pd = None


def has_pandas():
    """Returns True iff pandas can be imported."""
    return pd is not None


def numpy_plugin(_, v):
    """Represents numpy arrays as strings "ndarray(hash='xxx')"."""
    if np is not None and hasher is not None:
        if isinstance(v, np.ndarray):
            return "ndarray(hash='%s')" % hasher(v)
    return None


def pandas_plugin(_, v):
    """Represents pandas objects as any of "DataFrame(hash='xxx')" or "Series(hash='xxx')"."""
    if pd is not None and hasher is not None:
        if isinstance(v, pd.DataFrame):
            return "DataFrame(hash='%s')" % hasher(v)
        elif isinstance(v, pd.Series):
            return "Series(hash='%s')" % hasher(v)
    return None


# --- Plugin management

class WhatamiPluginManager(object):
    """
    Examples
    --------
    >>> def float_plugin(_, v):
    ...     if isinstance(v, float):
    ...         return "'float=%g'" % v
    ...     return None
    >>> function_plugin in WhatamiPluginManager.plugins()
    True
    >>> float_plugin in WhatamiPluginManager.plugins()
    False
    >>> WhatamiPluginManager.insert(float_plugin)
    >>> float_plugin in WhatamiPluginManager.plugins()
    True
    >>> float_plugin == WhatamiPluginManager.plugins()[-3]
    True
    >>> WhatamiPluginManager.drop(float_plugin)
    >>> float_plugin in WhatamiPluginManager.plugins()
    False
    >>> WhatamiPluginManager.insert(float_plugin, before=None)
    >>> float_plugin == WhatamiPluginManager.plugins()[-1]
    True
    >>> WhatamiPluginManager.reset()
    >>> WhatamiPluginManager.insert(float_plugin)
    >>> WhatamiPluginManager.reset()
    >>> float_plugin in WhatamiPluginManager.plugins()
    False
    >>> WhatamiPluginManager.drop(float_plugin)
    Traceback (most recent call last):
    ...
    ValueError: cannot drop plugin float_plugin, not in plugins list
    >>> WhatamiPluginManager.insert(string_plugin)
    Traceback (most recent call last):
    ...
    ValueError: cannot insert plugin string_plugin, already in plugins list
    >>> WhatamiPluginManager.insert(float_plugin, before=string_plugin)
    >>> float_plugin in WhatamiPluginManager.plugins()
    True
    >>> WhatamiPluginManager.drop(float_plugin)
    >>> WhatamiPluginManager.insert(float_plugin, before=float_plugin)
    Traceback (most recent call last):
    ...
    ValueError: plugin to insert before (float_plugin) not in plugins list
    >>> WhatamiPluginManager.insert(float_plugin)
    >>> print(whatareyou(lambda x=0.7: x))
    lambda(x='float=0.7')
    >>> WhatamiPluginManager.reset()
    """

    DEFAULT_PLUGINS = (what_plugin,
                       whatable_plugin,
                       builtin_plugin,
                       property_plugin,
                       string_plugin,
                       tuple_plugin,
                       list_plugin,
                       dict_plugin,
                       set_plugin,
                       partial_plugin,
                       function_plugin,
                       pandas_plugin,
                       numpy_plugin,
                       anyobject0x_plugin,
                       anyobject_plugin)

    PLUGINS = DEFAULT_PLUGINS

    @classmethod
    def plugins(cls):
        """Returns a tuple with the currently considered plugins."""
        return cls.PLUGINS

    @classmethod
    def reset(cls):
        """Makes the plugin list the default list."""
        cls.PLUGINS = cls.DEFAULT_PLUGINS

    @classmethod
    def drop(cls, plugin):
        """Removes a plugin from the list of plugins."""
        plugins = list(cls.PLUGINS)
        try:
            plugins.remove(plugin)
            cls.PLUGINS = tuple(plugins)
        except ValueError:
            raise ValueError('cannot drop plugin %s, not in plugins list' % plugin.__name__)

    @classmethod
    def insert(cls, plugin, before=anyobject0x_plugin):
        """Inserts a new plugin in the list of plugins used to generate strings for values in What configurations.

        Parameters
        ----------
        plugin : function (what, value) -> string
          A function that checks for value type and if it applies generates a string representing it,
          optionally using the information on What instance what.

        before : function, default anyobject0x_plugin
          A plugin already registered in the list. Note that the last two plugins capture most objects,
          so usually plugins need to be inserted at least before them.
          If None, the plugin is inserted at the end of the list.

        Raises
        ------
        ValueError if plugin is already in the list of registered plugins of if before is not in the list.
        """
        if plugin in cls.PLUGINS:
            raise ValueError('cannot insert plugin %s, already in plugins list' % plugin.__name__)
        plugins = list(cls.PLUGINS)
        if before is None:
            plugins.append(plugin)
        else:
            try:
                index = plugins.index(before)
                plugins.insert(index, plugin)
            except ValueError:
                raise ValueError('plugin to insert before (%s) not in plugins list' % before.__name__)
        cls.PLUGINS = tuple(plugins)



# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import

from ..what import whatable
from ..registry import WhatamiRegistry

import pytest


@pytest.fixture()
def registry():
    return WhatamiRegistry()


WHAT2NICKS = ['rfc0']


@pytest.fixture(params=WHAT2NICKS, ids=WHAT2NICKS)
def what2nick(request):
    if request.param == 'rfc0':
        @whatable
        def rfc(x, y=3, z='xyz'):  # pragma: no cover
            return x, y, z
        whatid = "rfc(y=3,z='xyz')"
        nick = 'rfc0'
        return rfc, whatid, nick
    raise ValueError('unknown fixture "%s"' % request.param)  # pragma: no cover


def test_register(registry, what2nick):

    what, whatid, nick = what2nick

    # register and access
    with pytest.raises(ValueError):
        registry.register(what, None)
    with pytest.raises(ValueError):
        registry.register(None, nick)
    assert what == registry.register(what=what, nickname=nick)
    assert nick == registry.nick_or_id(what)
    assert whatid == registry.nick2id(nick)

    # do not allow double registration
    with pytest.raises(Exception) as excinfo:
        registry.register(what, 'YOUSHALLNOTPASS')
    assert 'id "%s" is already associated with nickname "%s", delete it before updating' % (whatid, nick) \
           in str(excinfo.value)

    with pytest.raises(Exception) as excinfo:
        registry.register('YOUSHALLNOTPASS()', nick)
    assert 'nickname "%s" is already associated with id "%s", delete it before updating' % (nick, whatid) \
           in str(excinfo.value)


def test_list(registry, what2nick):
    what, whatid, nick = what2nick
    registry.register(what=what, nickname=nick)
    assert registry.list() == [(nick, whatid)]


def test_remove(registry, what2nick):

    what, whatid, nick = what2nick

    # remove by nickname
    assert what == registry.register(what, nick)
    assert registry.list() == [(nick, whatid)]
    registry.remove(nickname=nick)
    assert registry.list() == []

    # remove by whatid
    assert what == registry.register(what, nick)
    assert registry.list() == [(nick, whatid)]
    registry.remove(what=what)
    assert registry.list() == []

    # remove by both need to match the entry
    assert what == registry.register(what, nick)
    assert registry.list() == [(nick, whatid)]
    with pytest.raises(ValueError):
        registry.remove(what=what, nickname='NO')
    assert registry.list() == [(nick, whatid)]
    with pytest.raises(ValueError):
        registry.remove(what='NO()', nickname=nick)

    # remove does not accept None parameters
    with pytest.raises(ValueError) as excinfo:
        registry.remove()
    assert 'at least one of nickname or what must be provided' == str(excinfo.value)


def test_reset(registry, what2nick):
    what, whatid, nick = what2nick
    assert what == registry.register(what, nick)
    assert registry.list() == [(nick, whatid)]
    registry.reset()
    assert registry.list() == []



# coding=utf-8
"""Whatable wrappers for scikit learn tools.

Using it is fairly simple and does not require changing anything that is already using sklearn,
but beware that we are monkey-patching sklearn BaseEstimator to add the what method.

Example
-------

>>> # Monkey-patch sklearn estimators
>>> whatamise_sklearn()
>>> # Use what() to retrieve the id of a model
>>> from sklearn.ensemble import RandomForestClassifier
>>> rfc = RandomForestClassifier(n_jobs=8, n_estimators=23)
>>> print(rfc.what().id())
rfc(bootstrap=True,class_weight=None,criterion='gini',max_depth=None,max_features='auto',max_leaf_nodes=None,min_samples_leaf=1,min_samples_split=2,min_weight_fraction_leaf=0.0,n_estimators=23,random_state=None,warm_start=False)
>>> print(rfc.what())
rfc(bootstrap=True,class_weight=None,criterion='gini',max_depth=None,max_features='auto',max_leaf_nodes=None,min_samples_leaf=1,min_samples_split=2,min_weight_fraction_leaf=0.0,n_estimators=23,n_jobs=8,oob_score=False,random_state=None,verbose=0,warm_start=False)

Implementation Notes
--------------------

scikit-learn, in the good tradition of machine learning libraries (e.g. good old weka)
already deals graciously with configurability. Almost all interesting tools in sklearn
(classifiers, regressors, clusterers, dimensionality-reducers and other preprocessors)
are classes inheriting from BaseEstimator:
  https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/base.py

This is the key contract for these subclasses:
  "All estimators should specify all the parameters that can be set at the class level
   in their __init__ as explicit keyword arguments (no *args or **kwargs)."

This allows the sklearn folks to precisely define a protocol for configuring, configuration
retrieval and cloning (along with other niceties like standard control of deprecation).

It also makes really easy to add a what() method to scikit estimators.
"""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
import inspect
import warnings

from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans
from sklearn.cross_decomposition import PLSRegression, PLSCanonical, CCA, PLSSVD
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor, ExtraTreesClassifier, ExtraTreesRegressor
from sklearn.ensemble.gradient_boosting import GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.gaussian_process import GaussianProcess
from sklearn.linear_model import Ridge, Lasso, ElasticNet, Lars, OrthogonalMatchingPursuit, \
    BayesianRidge, ARDRegression, LogisticRegression, SGDClassifier, SGDRegressor, Perceptron, LassoLars, \
    LinearRegression
from sklearn.naive_bayes import GaussianNB, MultinomialNB, BernoulliNB
from sklearn.neighbors import KNeighborsClassifier, RadiusNeighborsClassifier, KNeighborsRegressor, \
    RadiusNeighborsRegressor, NearestCentroid
from sklearn.pipeline import FeatureUnion
from sklearn.preprocessing import Normalizer
from sklearn.svm import SVC, NuSVC, LinearSVC, SVR, NuSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from ..what import What
from ..misc import all_subclasses


# ----- Manually check which parameters are not part of the default id (WIP)

_LAST_SKLEARN_CHECKED = '1.5.1'

_SKLRegistry = {}


class _SklearnEstimatorID(object):
    """
    Stores information about sklearn short names, non_id_params, notes and possibly others.

    N.B. it is better to include a non-existing non_id_param
    (innocuous, can help to support several sklearn versions)
    than to leave out existing non_id_params.
    """

    def __init__(self,
                 skclass,
                 short_name,
                 non_id_params=(),
                 notes=None):
        self.skclass = skclass
        self.short_name = short_name if short_name is not None else skclass.__name__
        self.non_id_params = non_id_params
        self.notes = notes


def _a(skclass, nickname, non_id_params=(), notes=None):
    """Adds a class to the SKLRegistry."""
    _SKLRegistry[skclass.__name__] = _SklearnEstimatorID(skclass, nickname, non_id_params, notes)


# Ensembles
_a(RandomForestClassifier, 'rfc', ('verbose', 'n_jobs', 'oob_score'),
   'oob_score used to change the internal representation of the model (recheck)')
_a(RandomForestRegressor, 'rfr', ('verbose', 'n_jobs', 'oob_score'),
   'Same non-ids as RFC')
_a(ExtraTreesClassifier, 'etc', ('n_jobs', 'verbose', 'oob_score'))
_a(ExtraTreesRegressor, 'etr', ('n_jobs', 'verbose', 'oob_score'))
_a(GradientBoostingClassifier, 'gbc', ('verbose',))
_a(GradientBoostingRegressor, 'gbr', ('verbose',))

# GLMs
_a(Ridge, 'ridge', ('copy_X',))
_a(Lasso, 'lasso', ('copy_X',),
   'we need to take care of "precompute" when we pass the Gram matrix, should use a hash of the array')
_a(ElasticNet, 'elnet', ('copy_X',),
   'we need to take care of "precompute" when we pass the Gram matrix, should use a hash of the array')
_a(Lars, 'lars', ('verbose', 'copy_X', 'fit_path'),
   'we need to take care of "precompute" when we pass the Gram matrix, should use a hash of the array')
_a(LassoLars, 'lassolars', ('verbose', 'copy_X', 'fit_path'),
   'we need to take care of "precompute" when we pass the Gram matrix, should use a hash of the array')
_a(OrthogonalMatchingPursuit, 'omp', ('copy_X', 'copy_Xy', 'copy_Gram'),
   'we need to take care of "precompute" when we pass the Gram matrix, should use a hash of the array')
_a(BayesianRidge, 'bayridge', ('copy_X', 'verbose'))
_a(ARDRegression, 'ardr', ('copy_X', 'verbose'))
_a(LogisticRegression, 'logreg')
_a(SGDClassifier, 'sgdc', ('verbose', 'n_jobs'))
_a(SGDRegressor, 'sgdr', ('verbose',))
_a(Perceptron, 'perceptron', ('verbose', 'n_jobs'))
_a(LinearRegression, 'lr', ('copy_X',))

# SVMs
_a(SVC, 'svc', ('cache_size', 'verbose'))
_a(NuSVC, 'nusvc', ('cache_size', 'verbose'))
_a(LinearSVC, 'linsvc', ('verbose',))
_a(SVR, 'svr', ('cache_size', 'verbose'))
_a(NuSVR, 'nusvr', ('cache_size', 'verbose'))

# NaiveBayes
_a(GaussianNB, 'gnb')
_a(MultinomialNB, 'mnb')
_a(BernoulliNB, 'bnb')

# Decision Trees
_a(DecisionTreeClassifier, 'dtc')
_a(DecisionTreeRegressor, 'dtr')

# Nearest Neighbours
_a(KNeighborsClassifier, 'knc', ('warn_on_equidistant',))
_a(RadiusNeighborsClassifier, 'rnc', ('outlier_label',),
   'outlier_label is questionable, but it won\'t change the model')
_a(KNeighborsRegressor, 'knr', ('warn_on_equidistant',))
_a(RadiusNeighborsRegressor, 'rnr')
_a(NearestCentroid, 'nc')

# Partial Least Squares
_a(PLSRegression, 'plsr', ('copy',))
_a(PLSCanonical, 'plscan', ('copy',))
_a(CCA, 'cca', ('copy',))
_a(PLSSVD, 'plssvd')

# Gaussian Processes
_a(GaussianProcess, 'gp', ('storage_mode', 'verbose'))

# Clusterers
_a(KMeans, None, ('n_jobs', 'copy_x'))

# Preprocessing
_a(Normalizer, None, ('copy',))

# Pipelines
_a(FeatureUnion, 'f_union', ('n_jobs',))


# Make the invese map: short_name -> long_name
_SKLShort2Long = dict((v.short_name, k) for k, v in _SKLRegistry.items())


# ----- Monkey-patching for whatability


def _what_for_sklearn(x):
    """Returns a What configuration for the scikit learn estimator x."""
    name = x.__class__.__name__
    configuration_dict = x.get_params(deep=False)   # N.B. we take care of recursing ourselves
    pinfo = _SKLRegistry.get(name, None)
    if pinfo is not None:
        return What(pinfo.short_name,
                    conf=configuration_dict,
                    non_id_keys=pinfo.non_id_params)
    return What(name, conf=configuration_dict)


def whatamise_sklearn(check=False, log=False):
    """Monkey-patches sklearn.base.BaseEstimator to make sklearn estimators whatable.

    Parameters
    ----------
    check: boolean, default True
        If True, call check_all_monkeypatched() after getting right BaseEstimator.what
    """
    if not hasattr(BaseEstimator, 'what'):
        BaseEstimator.what = _what_for_sklearn
    if check:
        _check_all_monkeypatched()
    if log:
        print('Scikit-Learn estimators now are whatable')


def _check_all_monkeypatched():
    """Double-checks that instances sklearn estimators have acquired the proper "what" method.
    Raises an assertion error if it is not the case.
    """

    # Make sure we have added what to sklearn stuff
    whatamise_sklearn(check=False)

    # Trick to force python to populate part of the BaseEstimator hierarchy
    from sklearn.ensemble.forest import RandomForestClassifier
    assert BaseEstimator.__subclasscheck__(RandomForestClassifier)
    from sklearn.cluster import KMeans
    assert BaseEstimator.__subclasscheck__(KMeans)
    from sklearn.feature_extraction import DictVectorizer
    assert BaseEstimator.__subclasscheck__(DictVectorizer)
    from sklearn.decomposition import KernelPCA
    assert BaseEstimator.__subclasscheck__(KernelPCA)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        for cls in all_subclasses(BaseEstimator):
            if not inspect.isabstract(cls):
                try:
                    obj = cls()
                    assert hasattr(obj, 'what'), cls.__name__
                    assert isinstance(obj.what(), What), cls.__name__
                except TypeError:
                    pass
    return True


####
#
# Notes:
#  - As usual, when rng (random_state) is a None or a custom rng, identity is not really defined.
#    Maybe we could warn to always use an int random_state when this is detected (in what_for_sklearn).
#
#  - This is too aggressive on patching, but creating a lighter version by getting a plugin to whatareyou
#    that realizes about sklearn classes and applies ignores wisely is very easy to do.
#
####



# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
import hashlib

from future.utils import PY3

from ..what import What
from ..whatutils import id2what, what2id
from .fixtures import *


def test_configuration_nonids():

    # Non-ids
    c1 = What('tc',
              {'p1': 1, 'p2': 2, 'p3': 3, 'verbose': True, 'n_jobs': None},
              non_id_keys=('verbose', 'n_jobs'))
    assert c1.id() == 'tc(p1=1,p2=2,p3=3)'
    assert c1.id(nonids_too=True) == 'tc(n_jobs=None,p1=1,p2=2,p3=3,verbose=True)'

    with pytest.raises(Exception) as excinfo:
        What('tc',
             {'p1': 1, 'p2': 2, 'p3': 3, 'verbose': True, 'n_jobs': None},
             non_id_keys=str)
    assert str(excinfo.value) == 'non_ids must be None or an iterable'


def test_whatid():
    assert what2id(None) is None
    assert what2id('Myself') == 'Myself'
    assert what2id(int) == 'int()'  # correct behavior?


def test_non_id_keys(c3):
    # non-id keys
    config_c3 = c3.what()
    assert config_c3.id() == "C3(c1=C1(length=1,p1='blah',p2='bleh')," \
                             "c2=C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne'))"
    assert config_c3.id(nonids_too=True) == "C3(c1=C1(length=1,p1='blah',p2='bleh')," \
                                            "c2=C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne')," \
                                            "irrelevant=True)"
    c3id = "C3(c1=C1(length=1,p1='blah',p2='bleh'),c2=C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne'))"
    assert config_c3.id(nonids_too=False) == c3id
    sha1 = hashlib.sha1(c3id.encode('utf-8')).hexdigest()
    assert config_c3.id(maxlength=1) == sha1


def test_what_str_magic(c1, c2, c3):
    assert str(c1.what()) == "C1(length=1,p1='blah',p2='bleh')"
    assert str(c2.what()) == "C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne')"
    assert str(c3.what()) == "C3(c1=C1(length=1,p1='blah',p2='bleh')," \
                             "c2=C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne'),irrelevant=True)"


def test_what_repr_magic(c1):
    empty_dict_repr = 'set()' if PY3 else 'set([])'
    result = repr(c1.what()).replace("u'", "'")
    assert "What('C1', {" in result
    assert "'p2': 'bleh'" in result
    assert "'length': 1" in result
    assert "'p1': 'blah'" in result
    assert ("}, " + empty_dict_repr) in result
    # tests with c2 and c3 will fail because of pytest magic, rewrite without fixtures


def test_what_getitem_magic(c1, c2, c3):
    assert c1.what()['length'] == 1
    assert c1.what()['p1'] == 'blah'
    assert c2.what()['c1', 'p1'] == 'blah'
    assert c3.what()['c2', 'c1', 'p1'] == 'blah'
    with pytest.raises(KeyError):
        assert c1.what()['not_there']


def test_what_eq_magic(c1, c2, c3):
    assert c1.what() == c1.what()
    assert c1.what() != c2.what()
    assert c1.what() != c3.what()
    assert c2.what() == c2.what()
    assert c2.what() != c1.what()
    assert c2.what() != c3.what()
    assert c3.what() == c3.what()
    assert c3.what() != c1.what()
    assert c3.what() != c2.what()


def test_whatable_functions(c1):
    def identity(x):
        return x

    # Functions
    c1.p1 = identity
    assert c1.what().id() == "C1(length=1,p1=identity(),p2='bleh')"
    assert c1.p1(1) == 1


def test_list_parameters(c1):
    @whatable
    class WhatableWithList(object):
        def __init__(self):
            self.list = [c1, 'second']
    w = WhatableWithList()
    assert w.what().id() == "WhatableWithList(list=[%s,'second'])" % c1.what().id()
    w.list = []
    assert w.what().id() == "WhatableWithList(list=[])"


def test_what_keys():
    what = What('tom', conf={'a': 3, 'b': 'z', 'c': 3.14}, non_id_keys=['b'])
    assert what.keys() == ['a', 'c']
    assert what.keys(non_ids_too=True) == ['a', 'b', 'c']


def test_what_values():
    what = What('tom', conf={'a': 3, 'b': 'z', 'c': 3.14}, non_id_keys=['b'])
    assert what.values() == [3, 3.14]
    assert what.values(non_ids_too=True) == [3, 'z', 3.14]


def test_what_positional_ids():
    what = What('tom', conf={'a': 3, 'b': 'z', 'c': 3.14}, non_id_keys=['b'])
    assert what.positional_id() == 'tom(3,3.14)'
    assert what.positional_id(non_ids_too=True) == "tom(3,'z',3.14)"
    assert what.positional_id(name='jerry', non_ids_too=True) == "jerry(3,'z',3.14)"


def test_tuple_parameters(c1):
    @whatable
    class WhatableWithTuple(object):
        def __init__(self):
            self.tuple = ('first', c1)
    w = WhatableWithTuple()
    assert w.what().id() == "WhatableWithTuple(tuple=('first',%s))" % c1.what().id()
    w.tuple = ()
    assert w.what().id() == "WhatableWithTuple(tuple=())"


def test_dict_parameters(c1):
    @whatable
    class WhatableWithDict(object):
        def __init__(self):
            self.dict = {'c1': c1, 'two': 2}
    w = WhatableWithDict()
    assert w.what().id() == "WhatableWithDict(dict={'c1':C1(length=1,p1='blah',p2='bleh'),'two':2})"
    w.dict = {}
    assert w.what().id() == "WhatableWithDict(dict={})"


def test_set_parameters(c1):
    @whatable
    class WhatableWithSet(object):
        def __init__(self):
            self.set = {c1, 2}
    w = WhatableWithSet()
    assert w.what().id() == "WhatableWithSet(set={2,C1(length=1,p1='blah',p2='bleh')})"
    # whatables must have no memory of configuration
    w.set = set()
    assert w.what().id() == "WhatableWithSet(set=set())"


def test_lamda_id():

    def norm(x, y=3, normal=lambda x, ly=33: x + ly):  # pragma: no cover
        return normal(x) * x + y

    assert what2id(norm) == 'norm(normal=lambda(ly=33),y=3)'
    assert id2what(what2id(norm)) == What('norm', {'normal': What('lambda', {'ly': 33}), 'y': 3})


def test_what_copy(c1, c2, c3):
    for c in (c1, c2, c3):
        what = c.what()
        assert what == what.copy()


def test_numpy_plugin(array):

    array, array_hash2, array_hash3 = array
    array_hash = array_hash2 if not PY3 else array_hash3

    # joblib hash has changed?
    from whatami.plugins import hasher
    assert array_hash == hasher(array)

    @whatable
    def lpp(adjacency=array):  # pragma: no cover
        return adjacency

    assert lpp.what().id() == "lpp(adjacency=ndarray(hash='%s'))" % array_hash


def test_pandas_plugin(df):

    df, df_hash2, df_hash3 = df
    df_hash = df_hash2 if not PY3 else df_hash3
    name = df.__class__.__name__

    # check for changes in joblib hashing
    from whatami.plugins import hasher
    assert df_hash == hasher(df)

    # check for proper string generation
    @whatable
    def lpp(adjacency=df):  # pragma: no cover
        return adjacency

    assert lpp.what().id() == "lpp(adjacency=%s(hash='%s'))" % (name, df_hash)



# coding=utf-8
"""Unobtrusive object (self-)identification for python.

*whatami* strives to abstract configurability and experiment identifiability in a convenient way,
by allowing each object/computation to provide a string uniquelly and consistently self-identifying
itself.

It works this way:

  - Objects provide their own ids based on "parameter=value" dictionaries.
    They do so by returning an instance of the `What` class from a method named `what()`.
    What objects have in turn a method named `id()` providing reasonable strings
    to describe the object (computation).

  - whatami also provides a `whatable` decorator. It can be
    used to provide automatic creation of `What` objects from
    introspecting the object attributes or functions default parameters.

  - whatami automatically generated id strings tend to be long and therefore not human-friendly.
    A parser and transformer are provided to easily manipulate these strings

Examples
--------

>>> # Objects of this class provide a configuration (What object)
>>> class DuckedConfigurable(object):
...      def __init__(self, quantity, name, company=None, verbose=True):
...          self.quantity = quantity
...          self.name = name
...          self.company = company
...          self.verbose = verbose
...
...      def what(self):
...          return What('ducked', {'quantity': self.quantity, 'name': self.name, 'company': self.company})
>>>
>>>
>>> duckedc = DuckedConfigurable(33, 'salty-lollypops', verbose=False)
>>> # The configuration id string sorts by key alphanumeric order, helping id consistency
>>> print(duckedc.what().id())
ducked(company=None,name='salty-lollypops',quantity=33)
>>> # Using the whatable decorator makes objects gain a what() method
>>> # In this case, what() is infered automatically
>>> @whatable
... class Company(object):
...     def __init__(self, name, city, verbose=True):
...          self.name = name
...          self.city = city
...          self._verbose = verbose  # not part of config
...          self.social_reason_ = '%s S.A., %s' % (name, city)  # not part of config
>>> cc = Company(name='Chupa Chups', city='Barcelona')
>>> print(cc.what().id())
Company(city='Barcelona',name='Chupa Chups')
>>> # Ultimately, we can nest whatables...
>>> duckedc = DuckedConfigurable(33, 'salty-lollypops', company=cc, verbose=False)
>>> print(duckedc.what().id())
ducked(company=Company(city='Barcelona',name='Chupa Chups'),name='salty-lollypops',quantity=33)
>>> # Also a function decorator is provided - use with caution
>>> @whatable
... def buy(company, price=2**32, currency='euro'):
...     return '%s is now mine for %g %s' % (company.name, price, currency)
>>> print(buy.what().id())
buy(currency='euro',price=4294967296)
"""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import print_function, unicode_literals, absolute_import
import hashlib
import inspect
from functools import partial, update_wrapper, WRAPPER_ASSIGNMENTS
import types

from future.utils import PY3

from .misc import callable2call, is_iterable, config_dict_for_object, extract_decorated_function_from_closure, trim_dict


class What(object):
    """Stores and manipulates object configuration.

    Configurations are just dictionaries {key: value} that can nest and have a name.

    This helper class allows to represent configurations as (reasonable, python-like) strings.

    Parameters
    ----------
    name : string
        The name of this configuration (e.g. "RandomForest").

    conf : dictionary
        The {key:value} property dictionary for this configuration.

    non_id_keys : iterable (usually of strings), default None
        A list of keys that should not be considered when generating ids.
        For example: "num_threads" or "verbose" should not change results when fitting a model.
        Keys here won't make it to the configuration string unless explicitly asked for
    """

    __slots__ = ('name', 'conf', 'non_id_keys')

    def __init__(self,
                 name,
                 conf,
                 non_id_keys=None):
        super(What, self).__init__()
        self.name = name
        self.conf = conf
        if non_id_keys is None:
            self.non_id_keys = set()
        elif is_iterable(non_id_keys):
            self.non_id_keys = set(non_id_keys)
        else:
            raise Exception('non_ids must be None or an iterable')

    def copy(self):
        """Returns a copy of this whatable object.

        N.B. The configuration dictionary is copy is shallow;
        side-effects might happen if changes are made to mutable values.
        """
        return What(name=self.name, conf=self.conf.copy(), non_id_keys=self.non_id_keys)

    def flatten(self, non_ids_too=False, collections_too=False, recursive=True):
        """Returns two lists: keys and values.
        - keys is a list of tuples, each tuple being able to address a parameter of this what
        - values is a list with the value corresponding to each corresponding key in the keys array

        Parameters
        ----------
        non_ids_too : boolean, default False
          If True, include also non id keys

        collections_too : boolean, default False
          If True, include also keys that address members of python lists, tuples and dictionaries

        recursive : boolean, default True
          If True, recurse into nested What and collections

        Examples
        --------
        >>> what = whatareyou(lambda x=1, y=(1,2,{None: 3}): None)
        >>> keys, values = what.flatten(collections_too=True)
        >>> keys
        ['x', 'y', ('y', 0), ('y', 1), ('y', 2), ('y', 2, None)]
        >>> what[keys[0]]
        1
        >>> what[keys[-1]]
        3
        >>> values[0] == what[keys[0]]
        True
        >>> values[-1] == what[keys[-1]]
        True
        """
        def flatten(what, flattened_keys, flattened_values, partial_k):
            if isinstance(what, What):
                if non_ids_too:
                    kvs = sorted(what.conf.items())
                else:
                    kvs = sorted(item for item in what.conf.items() if item[0] not in self.non_id_keys)
            elif isinstance(what, (list, tuple)) and collections_too:
                kvs = enumerate(what)
            elif isinstance(what, dict) and collections_too:
                kvs = sorted(what.items())
            else:
                kvs = ()
            for k, v in kvs:
                flattened_keys.append(partial_k + (k,) if 0 < len(partial_k) else k)
                flattened_values.append(v)
                if recursive:
                    flatten(v, flattened_keys, flattened_values, partial_k + (k,))
            return flattened_keys, flattened_values
        return flatten(self, [], [], ())

    def keys(self, non_ids_too=False, collections_too=False, recursive=True):
        """Returns a list with the keys in the configuration."""
        return self.flatten(collections_too=collections_too, non_ids_too=non_ids_too, recursive=recursive)[0]

    def values(self, non_ids_too=False, collections_too=False, recursive=True):
        """Returns a list with the keys in the configuration."""
        return self.flatten(collections_too=collections_too, non_ids_too=non_ids_too, recursive=recursive)[1]

    # ---- Magics

    def __eq__(self, other):
        """Two configurations are equal if they have the same name and parameters."""
        return hasattr(other, 'name') and self.name == other.name and \
            hasattr(other, 'conf') and self.conf == other.conf

    def __str__(self):
        """The default representation is the configuration string including non_ids keys."""
        return self.id(nonids_too=True)

    def __repr__(self):
        return '%s(%r, %r, %r)' % (self.__class__.__name__, self.name, self.conf, self.non_id_keys)

    def __getitem__(self, item):
        """Allow to retrieve configuration values using [] notations, recursively, whatami aware."""
        try:
            return self.conf[item]
        except KeyError:
            if isinstance(item, tuple):
                w = self.conf
                for key in item:
                    try:
                        w = w[key]
                    except TypeError:
                        w = w.what()[key]
                return w
            raise

    # ---- ID string generation

    def id(self, nonids_too=False, maxlength=0):
        """Returns the id string (unicode) of this configuration.

        Parameters
        ----------
        nonids_too : boolean, default False
          Non-ids keys are ignored if nonids_too is False.

        malength : int, default 0
          If the id length goes over maxlength, it gets replaced by its sha1.
          If <= 0, it is ignored and the full id string will be returned.
        """
        kvs = ','.join('%s=%s' % (k, self.build_string(v))
                       for k, v in sorted(self.conf.items())
                       if nonids_too or k not in self.non_id_keys)
        my_id = '%s(%s)' % (self.name, kvs)
        return self._trim_too_long(my_id, maxlength=maxlength)

    def positional_id(self, name=None, non_ids_too=False, maxlength=0):
        """Returns an id without parameter names, just values.

        Parameters
        ----------
        name : string, default None
          The name to use in the id string; if None, the name of self is used.

        non_ids_too : boolena, default False
          Whether to include the non-id values too.

        maxlength : int, default 0
          If the id length goes over maxlength, it gets replaced by its sha1.
          If <= 0, it is ignored and the full id string will be returned.
        """
        name = name if name is not None else self.name
        my_id = '%s(%s)' % (name, ','.join(map(self.build_string, self.values(non_ids_too=non_ids_too))))
        return self._trim_too_long(my_id, maxlength=maxlength)

    def build_string(self, v):
        """Returns the nested configuration string for a variety of value types."""
        from .plugins import WhatamiPluginManager
        for plugin in WhatamiPluginManager.plugins():
            string = plugin(self, v)
            if string is not None:
                return string

    @staticmethod
    def _trim_too_long(string, maxlength=0):
        """Returns the string or its sha1 if the string length is larger than maxlength."""
        if 0 < maxlength < len(string):
            return hashlib.sha1(string.encode('utf-8')).hexdigest()
        return string


def whatareyou(obj,
               # ID string building options
               non_id_keys=None,
               # Config-dict building options
               add_dict=True,
               add_slots=True,
               add_properties=True,
               # Keys ignoring options
               exclude_prefix='_',
               exclude_postfix='_',
               excludes=('what',)):
    """Returns a What configuration following the specified behavior for inferring the configuration and ignoring keys.

    The meaning of all the parameters can be found in `What`, `config_dict_for_object`
    and `trim_dict`, to which this function delegates.

    Examples
    --------
    >>> def mola(a, n=5):
    ...     print(a + n)
    >>> print(whatareyou(mola).id())
    mola(n=5)
    >>> from functools import partial
    >>> print(whatareyou(partial(mola, n=7)))
    mola(n=7)
    """
    try:
        name, cd = callable2call(obj)
    except ValueError:
        name = obj.__class__.__name__
        cd = config_dict_for_object(obj, add_dict=add_dict, add_slots=add_slots, add_properties=add_properties)
    return What(name=name,
                conf=trim_dict(cd, exclude_prefix=exclude_prefix, exclude_postfix=exclude_postfix, excludes=excludes),
                non_id_keys=non_id_keys)


def is_whatable(obj):
    """Whatable objects have a method what() that takes no parameters and return a What configuration.

    Examples
    --------
    # >>> wp = whatable(partial(str, a=3))
    # >>> is_whatable(wp)
    # True
    # >>> is_whatable(str)
    # False
    >>> @whatable
    ... class WO(object):
    ...     def __init__(self):
    ...         self.a = 3
    >>> # The class is whatable...
    >>> is_whatable(WO)
    True
    >>> # ...so they are the instances
    >>> is_whatable(WO())
    True
    """
    try:
        what_method = obj.what
        selfbind = '__self__' if PY3 else 'im_self'
        if getattr(what_method, selfbind, None) is None:
            # Unbounded method, so this comes from a class
            if hasattr(what_method, 'whatami'):
                return True
            raise Exception('Cannot infer return type for unbound method what, '
                            'please pass a %r instance instead of the class' % obj)
        return isinstance(what_method(), What)
    except:
        return False


def whatable(obj=None,
             whatfunc=None,
             force_flag_as_whatami=False,
             # ID string building options
             non_id_keys=None,
             # Config-dict building options
             add_dict=True,
             add_slots=True,
             add_properties=True,
             exclude_prefix='_',
             exclude_postfix='_',
             excludes=('what',),
             # Other
             modify_func_inplace=False):
    """Decorates an object (also classes) to add a "what()" method.

    When decorating a callable (function, partial...), a brand new, equivalent callable will be
    returned (thus leaving the original intact). In this case, "what" provides safe ids
    only for results obtained when the function is called with its default parameters.
    This is useful in limited cases, for example, if we have a partial to fix all parameters
    but the data.

    When decorating non-callable objects or classes, this function adds a method "what"
    that respects all the {add_dict, add_slots, add_properties, exclude_prefix, exclude_postfix
    and excludes} as per `config_dict_for_object` and `trim_dict`.

    All the parameters are ignored if `whatfunc` is provided. It must be a function returning
    `whatfunc` should be a function accepting one object and must return another function
    that should return the relevant `What` object when called.

    Returns
    -------
    obj with a "what" method (or a wrapper function in case obj is originally a function)

    Examples
    --------
    >>> def normalize(x, mean=3, std=2):
    ...     return (x - mean) / std
    >>> cnormalize = whatable(normalize)
    >>> print(cnormalize.what().id())
    normalize(mean=3,std=2)
    >>> print(cnormalize.__name__)
    normalize
    >>> int(cnormalize(5))
    1
    >>> hasattr(normalize, 'what')
    False
    >>> @whatable
    ... def thunk(x, name='hi'):
    ...     print(x, name)
    >>> print(thunk.what().id())
    thunk(name='hi')
    >>> from future.moves.collections import UserDict
    >>> ud = whatable(UserDict())
    >>> is_whatable(ud)
    True
    >>> print(ud.what().id())
    UserDict(data={})
    >>> @whatable(add_properties=True)
    ... class WhatableWithProps(object):
    ...     def __init__(self):
    ...         super(WhatableWithProps, self).__init__()
    ...         self.a = 3
    ...         self._b = 2
    ...         self._c = 1
    ...     @property
    ...     def d(self):
    ...         return 0
    >>> wwp = WhatableWithProps()
    >>> is_whatable(wwp)
    True
    >>> print(wwp.what().id())
    WhatableWithProps(a=3,d=0)
    >>> wwp = whatable(wwp, add_dict=False, add_properties=True)
    >>> print(wwp.what().id())
    WhatableWithProps(d=0)
    """

    # class decorator
    if obj is None:
        return partial(whatable,
                       whatfunc=whatfunc,
                       force_flag_as_whatami=force_flag_as_whatami,
                       # ID string building options
                       non_id_keys=non_id_keys,
                       # Config-dict building options
                       add_dict=add_dict,
                       add_slots=add_slots,
                       add_properties=add_properties,
                       # Keys ignoring options
                       exclude_prefix=exclude_prefix,
                       exclude_postfix=exclude_postfix,
                       excludes=excludes,
                       modify_func_inplace=modify_func_inplace)

    # function decorator
    if inspect.isfunction(obj) or isinstance(obj, partial):

        if not modify_func_inplace:
            # Do not modify func inplace
            def whatablefunc(*args, **kwargs):
                return obj(*args, **kwargs)

            # Wrapper to get proper '__name__', '__doc__' and '__module__' when present
            # "wraps" won't work for partials or lambdas on python 2.x.
            # See: http://bugs.python.org/issue3445

            update_in_wrapper = [method for method in WRAPPER_ASSIGNMENTS if hasattr(obj, method)]
            if len(update_in_wrapper):
                whatablefunc = update_wrapper(wrapper=whatablefunc,
                                              wrapped=obj,
                                              assigned=update_in_wrapper)
        else:
            whatablefunc = obj

        # Adds what method
        if whatfunc is None:
            name, config_dict = callable2call(obj, closure_extractor=extract_decorated_function_from_closure)
            whatablefunc.what = partial(What, name=name, conf=config_dict, non_id_keys=excludes)
        else:
            whatablefunc.what = partial(whatfunc, whatablefunc)
        whatablefunc.what.whatami = True

        return whatablefunc

    if inspect.isbuiltin(obj):
        raise TypeError('builtins cannot be whatamised')

    # At the moment we just monkey-patch the object
    if hasattr(obj, 'what') and not is_whatable(obj):
        if force_flag_as_whatami:
            # mark this method as a whatami method
            # http://legacy.python.org/dev/peps/pep-0232/
            # http://stackoverflow.com/questions/23345005/does-pep-232-also-works-for-instance-methods
            to_annotate = obj.what.__func__ if not PY3 else obj.what
            to_annotate.whatami = True
            return obj
        else:
            raise Exception('object already has an attribute what, and is not a whatami what, if you know what I mean')

    try:
        def whatablefunc(self):
            return whatareyou(self,
                              non_id_keys=non_id_keys,
                              add_dict=add_dict,
                              add_slots=add_slots,
                              add_properties=add_properties,
                              exclude_prefix=exclude_prefix,
                              exclude_postfix=exclude_postfix,
                              excludes=excludes)
        whatablefunc = whatfunc if whatfunc is not None else whatablefunc
        whatablefunc.whatami = True
        if inspect.isclass(obj):
            obj.what = whatablefunc
        else:
            obj.what = types.MethodType(whatablefunc, obj)
        return obj
    except:
        raise Exception('cannot whatamise %s' % type(obj))



# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
from future.builtins import str
from sklearn.cluster import KMeans
from sklearn.ensemble import RandomForestClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer

from ..what_sklearn import whatamise_sklearn, _check_all_monkeypatched

whatamise_sklearn(check=True, log=True)


def test_monkeypatch():
    assert _check_all_monkeypatched()


def test_non_ids():
    rfc = RandomForestClassifier()
    assert 'n_jobs' not in rfc.what().id()
    assert 'n_jobs' in str(rfc.what())


def test_pipeline():
    norm = Normalizer(norm='l1')
    norm_id = norm.what().id()
    assert norm_id == "Normalizer(norm='l1')"
    kmeans = KMeans(n_clusters=12)
    kmeans_id = kmeans.what().id()
    assert kmeans_id == \
        "KMeans(init='k-means++',max_iter=300,n_clusters=12,n_init=10," \
        "precompute_distances='auto',random_state=None,tol=0.0001,verbose=0)"
    pipeline_id = Pipeline((('norm', norm), ('kmeans', kmeans))).what().id()
    assert pipeline_id == "Pipeline(steps=[('norm',%s),('kmeans',%s)])" % (norm_id, kmeans_id)



# coding=utf-8
"""Tests various functions in whatutils."""
from functools import partial
from ..whatutils import whatid2columns
from .fixtures import *


def test_whatid2columns(df_with_whatid):
    """Tests extracting values from whatid strings in a pandas dataframe column into their own columns.

    Uses df_with_whatid fixtures, that must provide a dataframe with:
      - A column "whatid" with the whatami ids
      - The rest of the columns must be the expectations for the extracted values, named as:
        - key: for top level keys
        - key1_key2_key3: for recursive keys
    """
    # Use pandas for testing
    from ..plugins import pd
    assert_series_equal = partial(pd.util.testing.assert_series_equal, check_names=False)
    whatid_column = 'whatid'
    expected_value_columns = sorted(set(df_with_whatid.columns) - {whatid_column})
    toplevel_values = [c for c in expected_value_columns if '_' not in c]
    recursive_values = [c.split('_') for c in expected_value_columns if '_' in c]
    # add without name change
    edf = whatid2columns(df_with_whatid, whatid_column, inplace=False)
    assert edf is not df_with_whatid
    for col in toplevel_values:
        assert_series_equal(df_with_whatid[col], edf[col])
    # recursive values
    edf = whatid2columns(df_with_whatid, whatid_column, recursive_values, prefix='pre-', inplace=False)
    assert edf is not df_with_whatid
    for col in recursive_values:
        colname = '_'.join(col)
        assert_series_equal(df_with_whatid[colname], edf['pre-' + colname])
    # add with prefix
    edf = whatid2columns(df_with_whatid, whatid_column, prefix='pre-', inplace=False)
    for col in toplevel_values:
        assert_series_equal(df_with_whatid[col], edf['pre-' + col])
    # add with postfix
    edf = whatid2columns(df_with_whatid, whatid_column, postfix='-post', inplace=False)
    for col in toplevel_values:
        assert_series_equal(df_with_whatid[col], edf[col + '-post'])
    # only some columns
    edf = whatid2columns(df_with_whatid, whatid_column,
                         columns=toplevel_values[:1],
                         prefix='pre-', postfix='-post', inplace=False)
    assert_series_equal(df_with_whatid[toplevel_values[0]], edf['pre-' + toplevel_values[0] + '-post'])
    for col in toplevel_values[1:]:
        assert ('pre-' + col + '-post') not in edf.columns
    # add inplace
    edf = whatid2columns(df_with_whatid, whatid_column, prefix='pre-', postfix='-post', inplace=True)
    assert edf is df_with_whatid
    for col in toplevel_values:
        assert_series_equal(df_with_whatid[col], df_with_whatid['pre-' + col + '-post'])



# coding=utf-8
"""Tests that id strings generated using introspection are valid."""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import

from ..what import whatable, What
from ..parsers import parse_whatid

import pytest


TEST_CASES = ['simple', 'collections', 'nested']


@pytest.fixture(params=TEST_CASES, ids=TEST_CASES)
def whatable2spectations(request):

    # Simple whatable
    if request.param == 'simple':
        @whatable
        def rfc(X, y, n_trees=10, split=2, depth=None, criterion='gini', alpha=1.34):  # pragma: no cover
            return X, y, n_trees, split, depth, criterion, alpha
        expected_id = "rfc(alpha=1.34,criterion='gini',depth=None,n_trees=10,split=2)"
        expected_what = What(name='rfc',
                             conf={'n_trees': 10,
                                   'split': 2,
                                   'depth': None,
                                   'criterion': 'gini',
                                   'alpha': 1.34})
        return rfc, expected_id, expected_what

    # Whatable with collections
    elif request.param == 'collections':

        a_dict = {'a': 2, None: 'c'}
        a_list = ['l', None, 3.2]
        an_empty_set = set()
        a_set = {'a', 3}
        a_tuple = (33, a_list, a_dict, an_empty_set, a_set)

        @whatable
        def rfc(X, y, t=a_tuple):    # pragma: no cover
            return X, y, t
        expected_id = "rfc(t=(33,['l',None,3.2],{'a':2,None:'c'},set(),{'a',3}))"
        expected_what = What(name='rfc', conf={'t': a_tuple})
        return rfc, expected_id, expected_what

    # Whatable with nested whatables
    elif request.param == 'nested':

        a_dict = {'a': 2, None: 'c'}
        a_list = ['l', None, 3.2]
        a_tuple = (33, a_list, a_dict)

        @whatable
        def rfc(X, y, t=a_tuple):  # pragma: no cover
            return X, y, t

        @whatable
        def nests(nested=rfc):  # pragma: no cover
            return nested
        expected_id = "nests(nested=rfc(t=(33,['l',None,3.2],{'a':2,None:'c'})))"
        expected_what = What(name='nests',
                             conf={'nested': What(name='rfc', conf={'t': (33, ['l', None, 3.2], {None: 'c', 'a': 2})})})
        return nests, expected_id, expected_what

    raise ValueError('Cannot find test case %s' % request.param)    # pragma: no cover


def test_roundtrip(whatable2spectations):
    a_whatable, expected_idstring, expected_what = whatable2spectations
    assert expected_idstring == a_whatable.what().id()
    what = parse_whatid(expected_idstring)
    assert expected_what.name == what.name
    assert expected_what.conf == what.conf



# coding=utf-8
"""Utilities for parsing and transforming whatami id strings."""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import print_function, unicode_literals, absolute_import
from arpeggio import ParserPython, Optional, ZeroOrMore, StrMatch, RegExMatch, EOF, PTNodeVisitor, visit_parse_tree


def build_whatami_parser(reduce_tree=False, debug=False):
    """
    Builds an arpeggio parser that accepts any valid whatami generated id string.

    (this function also describes the grammar of the language that What.id() strings should pertain to)

    Known limitations:
      - does not handle escaped single quoted in strings
       (for example, it will fail to parse "rfc(name='\'banyan\'')")

    Parameters
    ----------
    reduce_tree : boolean, default False
      Check `arpeggio.ParserPython` documentation

    debug : boolean, default False
      Check `arpeggio.ParserPython` documentation (N.B. will generate dot files in the running directory)

    Returns
    -------
    The arpeggio parser. Call `parser.parse` to generate the AST.
    """

    # Syntactic noise

    def list_sep():
        return StrMatch(',')

    # Basic types

    def an_id():
        # These account for valid python 2 identifiers. Python 3 allow unicode in identifiers.
        # See:
        #   http://stackoverflow.com/questions/5474008/regular-expression-to-confirm-whether-a-string-is-a-valid-identifier-in-python
        # When/if adapting to py3, that should be handled.
        # Note also that arpeggio does not allow unicode in regexps;
        # it should be easy to implement by just allowing arbitrary re flags in RegExMatch
        return RegExMatch(r'[_A-Za-z][_a-zA-Z0-9]*')

    def a_number():
        return RegExMatch('-?\d+((\.\d*)?((e|E)(\+|-)?\d+)?)?')

    def a_string():
        return StrMatch("'"), RegExMatch("[^']*"), StrMatch("'")

    def a_true():
        return StrMatch('True')

    def a_false():
        return StrMatch('False')

    def a_bool():
        return [a_true, a_false]

    def a_none():
        return StrMatch('None')

    # Collection types: lists, tuples, dictionaries

    def list_elements():
        return value, ZeroOrMore(list_sep, value)

    def a_list():
        return StrMatch('['), Optional(list_elements), StrMatch(']')

    def a_tuple():
        return StrMatch('('), Optional(list_elements), StrMatch(')')

    def dictkv():
        return value, StrMatch(':'), value

    def dict_elements():
        return dictkv, ZeroOrMore(list_sep, dictkv)

    def a_dict():
        return StrMatch('{'), Optional(dict_elements), StrMatch('}')

    def an_empty_set():
        return StrMatch('set()')

    def a_non_empty_set():
        return StrMatch('{'), Optional(list_elements), StrMatch('}')

    def a_set():
        return [an_empty_set, a_non_empty_set]

    # Key-values

    def value():
        return [a_none, a_bool, a_number, a_string, a_tuple, a_list, a_set, a_dict, whatami_id]

    def kv():
        return an_id, StrMatch('='), value

    def kvs():
        return kv, ZeroOrMore(list_sep, kv)

    # Top level

    def whatami_id():
        return an_id, StrMatch('('), Optional(kvs), StrMatch(')')

    def whatami_id_top():
        return whatami_id, EOF

    return ParserPython(whatami_id_top, reduce_tree=reduce_tree, debug=debug)


class WhatamiTreeVisitor(PTNodeVisitor):
    """A tree visitor for whatami id ASTs that returns a `whatami.What` object.

    Parameters
    ----------
    debug : boolean, default False
      Use debug output when visiting a tree

    Returns
    -------
    A `whatami.What` object inferred from the id string.
    """

    def __init__(self, debug=False):
        # N.B. these actions assume that syntactic noise is being ignored, therefore defaults=True
        super(WhatamiTreeVisitor, self).__init__(defaults=True, debug=debug)

    # --- Basic types

    @staticmethod
    def visit_an_id(node, _):
        return node.value

    @staticmethod
    def visit_a_number(node, _):
        try:
            return int(node.value)
        except ValueError:
            return float(node.value)

    @staticmethod
    def visit_a_string(_, children):
        return children[0]

    @staticmethod
    def visit_a_true(*_):
        return True

    @staticmethod
    def visit_a_false(*_):
        return False

    # None is the arpeggio marker for nodes to be ignored, we need this trick to defer None instantiation
    _NONE_NODE = '*-*NoNeNoDe*-*'

    @staticmethod
    def nonenode2none(x):
        return None if x is WhatamiTreeVisitor._NONE_NODE else x

    @staticmethod
    def visit_a_none(*_):
        return WhatamiTreeVisitor._NONE_NODE

    # --- Collection types: lists, tuples, dictionaries

    @staticmethod
    def visit_list_elements(_, children):
        return list(map(WhatamiTreeVisitor.nonenode2none, children))

    @staticmethod
    def visit_a_list(_, children):
        return children[0]

    @staticmethod
    def visit_a_tuple(_, children):
        return tuple(children[0])

    @staticmethod
    def visit_dictkv(_, children):
        return WhatamiTreeVisitor.nonenode2none(children[0]), WhatamiTreeVisitor.nonenode2none(children[1])

    @staticmethod
    def visit_dict_elements(_, children):
        return children

    @staticmethod
    def visit_a_dict(_, children):
        return dict(list(children[0]))

    @staticmethod
    def visit_an_empty_set(*_):
        return set()

    @staticmethod
    def visit_a_non_empty_set(_, children):
        return set(list(children[0]))

    # Key-values

    @staticmethod
    def visit_value(_, children):
        return children[0]

    @staticmethod
    def visit_kv(_, children):
        return children[0], WhatamiTreeVisitor.nonenode2none(children[1])

    @staticmethod
    def visit_kvs(_, children):
        return children

    # --- Top level

    @staticmethod
    def visit_whatami_id(_, children):
        from .what import What
        an_id = children[0]
        kvs = list(children[1]) if len(children) > 1 else []
        return What(an_id, dict(kvs))

    @staticmethod
    def visit_whatami_id_top(_, children):
        return children[0]


DEFAULT_WHATAMI_PARSER = build_whatami_parser()
DEFAULT_WHATAMI_VISITOR = WhatamiTreeVisitor()


def parse_whatid(id_string, parser=None, visitor=None):
    """
    Parses whatami id string into a pair (name, configuration).
    Makes a best effort to reconstruct python objects.

    Parameters
    ----------
    id_string : string
        The whatami id string to parse back.

    Returns
    -------
    A `whatami.What` object, containing name and conf (name is a string and configuration is a dictionary).

    Examples
    --------
    >>> what = parse_whatid('rfc(n_jobs=multiple(here=100))')
    >>> print(what.name)
    rfc
    >>> print(len(what.conf))
    1
    >>> print(what.conf['n_jobs'].conf['here'])
    100
    """
    global DEFAULT_WHATAMI_PARSER
    if parser is None:
        parser = DEFAULT_WHATAMI_PARSER
    if visitor is None:
        visitor = DEFAULT_WHATAMI_VISITOR
    try:
        return visit_parse_tree(parser.parse(id_string), visitor=visitor)
    except TypeError:
        # Remove this once arpeggio is released with this fix:
        # https://github.com/igordejanovic/Arpeggio/pull/21
        DEFAULT_WHATAMI_PARSER = build_whatami_parser()
        raise

# --- Maintenance for old whatami id strings


def build_oldwhatami_parser(reduce_tree=False, debug=False):

    # Syntactic noise

    def list_sep():
        return StrMatch(',')

    # Basic types

    def an_id():
        # These account for valid python 2 identifiers. Python 3 allow unicode in identifiers.
        # See:
        #   http://stackoverflow.com/questions/5474008/regular-expression-to-confirm-whether-a-string-is-a-valid-identifier-in-python
        # When/if adapting to py3, that should be handled.
        # Note also that arpeggio does not allow unicode in regexps;
        # it should be easy to implement by just allowing arbitrary re flags in RegExMatch
        return RegExMatch(r'[_A-Za-z][_a-zA-Z0-9]*')

    def a_number():
        return RegExMatch('-?\d+((\.\d*)?((e|E)(\+|-)?\d+)?)?')

    def a_string():
        return StrMatch("'"), RegExMatch("[^']*"), StrMatch("'")

    def a_true():
        return StrMatch('True')

    def a_false():
        return StrMatch('False')

    def a_bool():
        return [a_true, a_false]

    def a_none():
        return StrMatch('None')

    # Collection types: lists, tuples, dictionaries

    def list_elements():
        return value, ZeroOrMore(list_sep, value)

    def a_list():
        return StrMatch('['), Optional(list_elements), StrMatch(']')

    def a_tuple():
        return StrMatch('('), Optional(list_elements), StrMatch(')')

    def dictkv():
        return value, StrMatch(':'), value

    def dict_elements():
        return dictkv, ZeroOrMore(list_sep, dictkv)

    def a_dict():
        return StrMatch('{'), Optional(dict_elements), StrMatch('}')

    def an_empty_set():
        return StrMatch('set()')

    def a_non_empty_set():
        return StrMatch('{'), Optional(list_elements), StrMatch('}')

    def a_set():
        return [an_empty_set, a_non_empty_set]

    # Key-values

    def value():
        return [a_none, a_bool, a_number, a_string, a_tuple, a_list, a_set, a_dict, whatami_id]

    def kv():
        return an_id, StrMatch('='), value

    def kvs():
        return kv, ZeroOrMore('#', kv)

    # Top level

    def whatami_id():
        return [(an_id, StrMatch('#'), Optional(kvs)),
                (StrMatch('"'), an_id, StrMatch('#'), Optional(kvs), StrMatch('"'))]

    def whatami_id_top():
        return whatami_id, EOF

    return ParserPython(whatami_id_top, reduce_tree=reduce_tree, debug=debug)



# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
import arpeggio

from ..what import What
from ..parsers import parse_whatid

import pytest


# --- Parsing id strings

@pytest.mark.xfail(reason='known limitation, to fix or to document as permanent limitation')
def test_parse_whatid_within_string():  # pragma: no cover
    what = parse_whatid("rfc(name=''banyan'')")
    assert what.name == 'rfc'
    assert what.conf == {'name': "'banyan'"}


def test_parse_id_simple():
    # no parameters
    what = parse_whatid('rfc()')
    assert what.name == 'rfc'
    assert what.conf == {}

    # one parameter
    what = parse_whatid("rfc(n_jobs=4)")
    assert what.name == 'rfc'
    assert what.conf == {'n_jobs': 4}

    # multiple parameters
    what = parse_whatid("rfc(deep=True,gini=False,n_jobs=3.4,n_trees=100,seed='rng',splitter=None)")
    assert what.name == 'rfc'
    assert what.conf == {'n_jobs': 3.4, 'n_trees': 100, 'seed': 'rng', 'deep': True, 'splitter': None, 'gini': False}


def test_parse_id_lists():
    # lists become lists
    what = parse_whatid("rfc(splits=[1,7,'end'])")
    assert what.name == 'rfc'
    assert what.conf == {'splits': [1, 7, 'end']}
    # also with tuples
    what = parse_whatid("rfc(splits=(1,7,'end'))")
    assert what.name == 'rfc'
    assert what.conf == {'splits': (1, 7, 'end')}


def test_parse_id_dicts():
    what = parse_whatid("rfc(splits={1:7, None:'end', 'end':None, 'lst': [7, 'b', 1], 'd': {'nest': 2}})")
    assert what.name == 'rfc'
    assert what.conf == {'splits': {1: 7, None: 'end', 'end': None, 'lst': [7, 'b', 1], 'd': {'nest': 2}}}


def test_parse_id_sets():
    # with elements
    what = parse_whatid("rfc(splits={1, None, 'end'})")
    assert what.name == 'rfc'
    assert what.conf == {'splits': {1, None, 'end'}}
    # empty
    what = parse_whatid("rfc(splits=set())")
    assert what.name == 'rfc'
    assert what.conf == {'splits': set()}


def test_parse_id_spaces():
    what = parse_whatid("rfc(splits={  1:7,  None: 'end','end':None, 'lst': [7,'b', 1], 'd':  {'nest': 2}})")
    assert what.name == 'rfc'
    assert what.conf == {'splits': {1: 7, None: 'end', 'end': None, 'lst': [7, 'b', 1], 'd': {'nest': 2}}}


def test_parse_id_nested():
    what = parse_whatid('rfc(n_jobs=multiple(here=100))')
    assert what.name == 'rfc'
    assert what.conf == {'n_jobs': What('multiple', {'here': 100})}

    # machine learning inspired nasty nestness

    norm_id = 'Normalizer(norm=\'l1\')'
    kmeans_id = "KMeans(init='k-means++',max_iter=300,n_clusters=12,n_init=10," \
                "precompute_distances=True,random_state=None,tol=0.0001,verbose=0)"
    pipeline_id = "Pipeline(steps=[('norm', %s), ('clusterer', %s)])" % (norm_id, kmeans_id)
    what = parse_whatid(pipeline_id)
    assert what.name == 'Pipeline'
    assert what.conf == {
        'steps': [('norm', What('Normalizer', {'norm': 'l1'})),
                  ('clusterer', What('KMeans', {
                      'init': 'k-means++',
                      'max_iter': 300,
                      'n_clusters': 12,
                      'n_init': 10,
                      'precompute_distances': True,
                      'random_state': None,
                      'tol': 0.0001,
                      'verbose': 0
                  }))]}


def test_parse_id_wrong():

    # Configurations must not be empty
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('')

    # Configurations must have a name
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('()')

    # Configurations must have a configuration
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('wrong')

    # Values must have a key
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('rfc(5)')

    # Keys must have a value
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('rfc(x=, y=12)')

    # Configurations must have matching parenthesis
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('rfc(5')
    with pytest.raises(arpeggio.NoMatch):
        parse_whatid('rfc5)')


def test_arpeggio_resilience():
    with pytest.raises(TypeError):
        parse_whatid(map)
    what = parse_whatid("rfc(splits = {1, None, 'end'})")
    assert what.name == 'rfc'
    assert what.conf == {'splits': {1, None, 'end'}}



# coding=utf-8
from ..what import whatable, whatareyou
from ..plugins import has_numpy, has_joblib, has_pandas
import pytest


@pytest.fixture
def c1():
    """A simple whatable object."""
    @whatable
    class C1(object):
        def __init__(self, p1='blah', p2='bleh', length=1):
            super(C1, self).__init__()
            self.p1 = p1
            self.p2 = p2
            self.length = length
            self._p1p2 = p1 + p2
            self.p2p1_ = p2 + p1
    return C1()


@pytest.fixture
def c2(c1):
    """A whatable object with a nested whatable."""
    @whatable
    class C2(object):
        def __init__(self, name='roxanne', c1=c1):
            super(C2, self).__init__()
            self.name = name
            self.c1 = c1
    return C2()


@pytest.fixture
def c3(c1, c2):
    """A whatable object with nested whatables and irrelevant members."""

    @whatable(force_flag_as_whatami=True)
    class C3(object):
        def __init__(self, c1=c1, c2=c2, irrelevant=True):
            super(C3, self).__init__()
            self.c1 = c1
            self.c2 = c2
            self.irrelevant = irrelevant

        def what(self):
            return whatareyou(self, non_id_keys=('irrelevant',))
    return C3()


def numpy_skip(test):  # pragma: no cover
    """Skips a test if the numpy plugin is not available."""
    if not (has_numpy() and has_joblib()):
        return pytest.mark.skipif(test, reason='the numpy plugin requires both pandas and joblib')
    return test


@pytest.fixture(params=map(numpy_skip, ['a1', 'a2', 'a3', 'a4', 'a5', 'a6', 'a7']),
                ids=['a1', 'a2', 'a3', 'a4', 'a5', 'a6', 'a7'])
def array(request):
    """Hardcodes hashes, so we can detect hashing changes in joblib."""
    from ..plugins import np
    arrays = {
        # base array
        'a1': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]]),
               'a6bb4681650ec50fce0123412a78753e', 'cbded866f66a0fa6767b4e286c3552df'),
        # hash changes with dtype
        'a2': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]], dtype=np.bool),
               '82fe62950379505b6581df73d5a5bf2d', '5118dfbd9491eab8ce757c49b6fd06df'),
        'a3': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]], dtype=np.float),
               '1b5b918e0bae98539bb7aa886c791548', '7149c69cf4a5f85bd49e92496d5d2cb8'),
        # hash changes with shape and ndim
        'a4': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]]).reshape((1, 9)),
               'bc2afdb8b2d4ac89b5718105c554921b', '37c27fea094cf3eddf4b11e602955c2a'),
        'a5': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]], ndmin=3),
               '657a4d5a3a3e2190e21d1a06772b90fc', '33d23b13c6a0d41d9b6273ff4962f6c9'),
        # hash changes with stride/order/contiguity
        'a6': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]], order='F'),
               'fddb29315104f69723750835086584bf', '6465c08894edcc3d0d122b2fd0acb68f'),
        'a7': (np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]]).T,
               'fddb29315104f69723750835086584bf', '6465c08894edcc3d0d122b2fd0acb68f'),
    }
    return arrays[request.param]


def pandas_skip(test):  # pragma: no cover
    """Skips a test if the pandas plugin is not available."""
    if not (has_pandas() and has_joblib()):
        return pytest.mark.skipif(test, reason='the pandas plugin requires both pandas and joblib')
    return test


@pytest.fixture(params=map(pandas_skip, ['df1', 'df2', 'df3', 'df4', 's1', 's2']),
                ids=['df1', 'df2', 'df3', 'df4', 's1', 's2'])
def df(request):
    """Hardcodes hashes, so we can detect hashing changes in joblib."""
    from ..plugins import pd, np
    adjacency = np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]])
    dfs = {
        'df1': (pd.DataFrame(data=adjacency, columns=['x', 'y', 'z']),
                'fc5d5b43464052efbcd9e88dc0be4afd', '62171fe1114d5d961742dd95e6af37d7'),
        'df2': (pd.DataFrame(data=adjacency, columns=['xx', 'yy', 'zz']),
                'fe226e22261ff56d4cf019aa64942050', '139261e54b3ac2f2e39da6d497f6d0fd'),
        'df3': (pd.DataFrame(data=adjacency.T, columns=['x', 'y', 'z']),
                'f3d11416b1d5cb5da7753991941beb0d', 'fcd984c10ea9379faee471eedafee77b'),
        'df4': (pd.DataFrame(data=adjacency, columns=['x', 'y', 'z'], index=['r1', 'r2', 'r3']),
                '95fc40cb4635424c7ef1577d13183afa', 'c5234b1a362ef13c9c12b7b7444b8c85'),
        's1': (pd.Series(data=adjacency.ravel()),
               'e37122dc5f6320e9f12b413631056443', 'ee9729300f29a6917f30aa9e612ec67c'),
        's2': (pd.Series(data=adjacency.ravel(), index=list(range(len(adjacency.ravel()))))[::-1],
               'c0f4565b063599c6075ec6108cbca344', '74e14992d8587454d561b3194d11a984'),
    }
    return dfs[request.param]


@pytest.fixture(params=map(pandas_skip, ['dfw1']),
                ids=['dfw1'])
def df_with_whatid(request):
    """Fixtures to test whatid manipulations mixed with pandas dataframes.

    Provides dataframes with:
      - A column "whatid" with the whatami ids
      - The rest of the columns must be the expectations for the extracted values, named as:
        - key: for top level keys
        - key1_key2_key3: for recursive keys
    """
    from ..plugins import pd
    if request.param == 'dfw1':
        whatids = [
            "Blosc(cname='blosclz',level=5,shuffle=False)",
            "Blosc(cname='blosclz',level=6,shuffle=True)",
            "Blosc(cname='lz4hc',level=7,shuffle=True)",
            "Blosc(cname='lz4hc',level=8,shuffle=False)",
            "C2(c1=C1(length=1,p1='blah',p2='bleh'),name='roxanne')",
        ] * 4
        cnames = ['blosclz', 'blosclz', 'lz4hc', 'lz4hc', None] * 4
        levels = [5, 6, 7, 8, None] * 4
        shuffles = [False, True, True, False, None] * 4
        c1_lengths = [None, None, None, None, 1] * 4
        df = pd.DataFrame({'whatid': whatids,
                           'cname': cnames,
                           'level': levels,
                           'shuffle': shuffles,
                           'c1_length': c1_lengths})
        return df
    else:  # pragma: no cover
        raise ValueError('Unknown fixture %s' % request.param)



#!/usr/bin/env python2
# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='whatami',
    license='BSD 3 clause',
    description='Easily provide python objects with self-identification',
    long_description=open('README.rst').read().replace('|Build Status| |Coverage Status| |Scrutinizer Status|', ''),
    version='4.0.0',
    url='https://github.com/sdvillal/whatami',
    author='Santi Villalba',
    author_email='sdvillal@gmail.com',
    packages=['whatami',
              'whatami.tests',
              'whatami.wrappers',
              'whatami.wrappers.tests'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Operating System :: Unix',
    ],
    install_requires=['arpeggio>=1.0', 'future'],
    tests_require=['pytest', 'pytest-cov', 'pytest-pep8'],
    extras_require={
        'sklearn': ['scikit-learn'],
        'docs': ['Sphinx'],
    },
    platforms=['Any'],
)



# coding=utf-8

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
from datetime import datetime
import inspect
from time import strptime, mktime

from functools import partial
from future.utils import PY3

from ..what import whatable
from ..misc import callable2call, is_iterable, mlexp_info_helper

import pytest


def test_is_iterable():
    assert is_iterable([1, 2, 3]) is True
    assert is_iterable([]) is True
    assert is_iterable((1, 2, 3)) is True
    assert is_iterable(()) is True
    assert is_iterable(range(19)) is True
    assert is_iterable('abc') is True
    assert is_iterable(None) is False
    assert is_iterable(2) is False
    assert is_iterable(2.5) is False
    assert is_iterable(str) is False


def test_callable2call_partials():
    assert callable2call(partial(map)) == ('map', {})
    assert callable2call(partial(map, function=str)) == ('map', {'function': str})
    assert callable2call(partial(partial(map, function=str), iterable1=())) == ('map', {'function': str,
                                                                                        'iterable1': ()})
    with pytest.raises(Exception) as excinfo:
        callable2call(partial(test_callable2call_partials, function=str, f2=2, f3=3))
    assert 'keywords are not parameters of the function' in str(excinfo.value)

    with pytest.raises(Exception) as excinfo:
        callable2call(partial(test_callable2call_partials, 2, 3))
    assert 'There are too many positional arguments indicated ' in str(excinfo.value)

    def myf(x, y=3):  # pragma: no cover
        return x + y

    with pytest.raises(Exception) as excinfo:
        callable2call(partial(myf, 2, x=3))
    assert 'Some arguments are indicated both by position and name ' in str(excinfo.value)


def test_callable2call_functions():
    assert callable2call(test_callable2call_functions) == ('test_callable2call_functions', {})
    assert callable2call(partial) == ('partial', {})

    def f(x, y=3):
        return x + y

    assert callable2call(f) == ('f', {'y': 3})
    assert f(3) == 6


def test_callable2call_builtins():
    assert callable2call(sorted) == ('sorted', {})


def test_callable2call_wrongargs():
    with pytest.raises(Exception) as excinfo:
        callable2call('sorted')
    if PY3:  # pragma: no cover
        expected = 'Only callables (partials, functions, builtins...) are allowed, \'sorted\' is none of them'
    else:  # pragma: no cover
        expected = 'Only callables (partials, functions, builtins...) are allowed, u\'sorted\' is none of them'
    assert str(excinfo.value) == expected


def test_callable2call_lambdas():
    # Anonymous functions are a bit of a corner case
    assert callable2call(lambda x: x) == ('lambda', {})
    assert callable2call(lambda x=5: x) == ('lambda', {'x': 5})


def test_mlexp_info_helper():

    @whatable
    class TestDataset(object):
        def __init__(self):
            super(TestDataset, self).__init__()

    @whatable
    class Prepro(object):
        def __init__(self, lower=0, upper=1):
            super(Prepro, self).__init__()
            self.min = lower
            self.max = upper

    @whatable
    class PreproModel(object):
        def __init__(self, prepro=None, reg='l1', C=1.):
            super(PreproModel, self).__init__()
            self.prepro = prepro
            self.reg = reg
            self.C = C

    @whatable
    class CVEval(object):
        def __init__(self, num_folds=10, seed=0):
            super(CVEval, self).__init__()
            self.num_folds = num_folds
            self.seed = seed

    before = int(datetime.now().strftime("%s"))
    info = mlexp_info_helper(
        title='test',
        data_setup=TestDataset().what(),
        model_setup=PreproModel(prepro=Prepro(), reg='l2').what(),
        eval_setup=CVEval(num_folds=5, seed=2147483647).what(),
        exp_function=test_mlexp_info_helper,
        comments='comments4nothing',
        itime=None)
    assert info['title'] == 'test'
    assert info['data_setup'] == 'TestDataset()'
    assert info['model_setup'] == "PreproModel(C=1.0,prepro=Prepro(max=1,min=0),reg='l2')"
    assert info['eval_setup'] == 'CVEval(num_folds=5,seed=2147483647)'
    assert info['fsource'] == inspect.getsourcelines(test_mlexp_info_helper)
    assert info['comments'] == 'comments4nothing'
    recorded_time = mktime(strptime(info['date'], '%Y-%m-%d %H:%M:%S'))
    assert (recorded_time - before) < 2



# coding=utf-8
"""Simple implementations of central registries for whatami ids, nicknames, synonyms and contexts."""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import unicode_literals, absolute_import
from .whatutils import what2id


class WhatamiRegistry(object):
    """Bidirectional mapping, one-to-one, no persistence ATM (all should be in code), string to string (no thunks)."""

    def __init__(self, name='master'):
        super(WhatamiRegistry, self).__init__()
        self.name = name
        self._id2nick = {}
        self._nick2id = {}

    def register(self, what, nickname):
        """Registers the nickname for the id of what; returns `what` in the interest of fluent interfaces."""

        if what is None:
            raise ValueError('what cannot be None')
        if nickname is None:
            raise ValueError('nickname cannot be None')

        whatid = what2id(what)

        i2n, n2i = self._id2nick, self._nick2id

        # Ensure a one-to-one relationship; probably a nickname pointing to more than one id would be more useful
        if nickname in n2i and not n2i[nickname] == whatid:
            raise Exception('nickname "%s" is already associated with id "%s", delete it before updating' %
                            (nickname, n2i[nickname]))
        if whatid in i2n and not i2n[whatid] == nickname:
            raise Exception('id "%s" is already associated with nickname "%s", delete it before updating' %
                            (whatid, i2n[whatid]))

        # Add binding
        i2n[whatid] = nickname
        n2i[nickname] = whatid

        # Fluent
        return what

    def list(self):
        """Returns a sorted list of tuples (nick, id)."""
        return sorted(self._nick2id.items())

    def remove(self, nickname=None, what=None):
        """Removes the entry corresponding to nickname or what.
        If both are provided, they must correspond to a single entry.
        """
        whatid = None if what is None else what2id(what)
        if nickname is None and whatid is None:
            raise ValueError('at least one of nickname or what must be provided')
        elif nickname is None:
            nickname = self.id2nick(whatid)
        elif what is None:
            whatid = self.nick2id(nickname)
        else:
            old_whatid = self.nick2id(nickname)
            old_nick = self.id2nick(nickname)
            if old_whatid != whatid or old_nick != nickname:
                raise ValueError('both whatid "%s"=="%s" and nickname "%s"=="%s" identities must hold' %
                                 (whatid, old_whatid, nickname, old_nick))
        del self._nick2id[nickname]
        del self._id2nick[whatid]

    def nick2id(self, nickname):
        """Maps a nickname to the corrensponding id, returning None if the pair is not in the registry."""
        return self._nick2id.get(nickname, None)

    def id2nick(self, whatid):
        """Maps an id to the corrensponding nickname, returning None if the pair is not in the registry."""
        return self._id2nick.get(whatid, None)

    def nick_or_id(self, what):
        """Returns the nickname if it exists, otherwise it returns the id."""
        whatid = what2id(what)
        return self._id2nick.get(whatid, whatid)

    def reset(self):
        """Removes all entries in the registry."""
        self._id2nick = {}
        self._nick2id = {}



# coding=utf-8
"""Some utilities built on top of the `whatami.What`, `whatami.whatable` and `whatami.parse_whatid` machinery."""

# Authors: Santi Villalba <sdvillal@gmail.com>
# Licence: BSD 3 clause

from __future__ import print_function

import inspect
from itertools import chain

from operator import itemgetter
from arpeggio import NoMatch

from future.builtins import str as str3

from .parsers import parse_whatid, build_oldwhatami_parser
from .what import whatable, whatareyou, What, is_whatable


def what2id(obj):
    """Returns the configuration of obj as a string.

    Returns
    -------
      None if obj is None
      obj if obj is a string
      obj.what().id() if obj has a what() method returning a whatable
      obj.id() if obj has an id() method
      whatareyou(obj).id() otherwise
    """
    if obj is None:
        return None
    if isinstance(obj, str3):
        return obj
    try:
        return obj.what().id()
    except:
        try:
            return obj.id()
        except:
            return whatareyou(obj).id()


# An alias for whatami.parse_whatid
id2what = parse_whatid


def obj2what(obj,
             # ID string building options
             non_id_keys=None,
             # Config-dict building options
             add_dict=True,
             add_slots=True,
             add_properties=True,
             # Keys ignoring options
             exclude_prefix='_',
             exclude_postfix='_',
             excludes=('what',)):
    """Returns returns a best-guess for a suitable `What` for the obj:
     - `obj.what()` if obj is a `whatable`
     - otherwise invokes `whatareyou` on the object with the provided options.

    Examples
    --------
    >>> print(obj2what(whatable(id2what)).id())
    parse_whatid(parser=None,visitor=None)
    >>> print(obj2what(id2what).id())
    parse_whatid(parser=None,visitor=None)
    >>> print(obj2what(id2what, excludes=('parser',)).id())
    parse_whatid(visitor=None)
    """
    if is_whatable(obj):  # do not move this to whatareyou, or we face infinite recursion
        return obj.what()
    return whatareyou(obj,
                      non_id_keys=non_id_keys,
                      add_dict=add_dict,
                      add_slots=add_slots,
                      add_properties=add_properties,
                      # Keys ignoring options
                      exclude_prefix=exclude_prefix,
                      exclude_postfix=exclude_postfix,
                      excludes=excludes)


def whatvalues(what, keys=()):
    """Returns a tuple with the values assigned to keys in the the what.what() What object.
    Examples
    --------
    >>> whatvalues(obj2what(id2what), ('parser', 'visitor'))
    (None, None)
    >>> whatvalues(obj2what(id2what), 'parser')
    (None,)
    """
    if not isinstance(keys, tuple):
        keys = (keys,)
    return tuple(what[key] for key in keys)


def sort_whats(whats, *keys):
    """
    Sorts a list of What objects according to the value of some parameters.
    Examples
    --------
    >>> ids = ["Lagged(fex=distcorr(), lag=%d, response='acceleration', stimulus='force')" % lag
    ...        for lag in range(-2, 3)][::-1]
    >>> whats = list(map(id2what, ids))
    >>> whats_sorted, values = sort_whats(whats, 'lag')
    >>> [what['lag'] for what in whats]
    [2, 1, 0, -1, -2]
    >>> list(val for (val,) in values)
    [-2, -1, 0, 1, 2]
    >>> [what['lag'] for what in whats_sorted]
    [-2, -1, 0, 1, 2]
    """
    values = [whatvalues(what, keys) for what in whats]
    return tuple(zip(*[(whatid, value) for value, whatid in sorted(zip(values, whats), key=itemgetter(0))]))


def sort_whatids(whatids, *keys):
    """
    Sorts a list of whatami identifiers according to the value of some parameters.

    Examples
    --------
    >>> ids = ["Lagged(fex=distcorr(), lag=%d, response='acceleration', stimulus='force')" % lag
    ...        for lag in range(-2, 3)][::-1]
    >>> ids_sorted, values = sort_whatids(ids, 'lag')
    >>> [what['lag'] for what in map(id2what, ids)]
    [2, 1, 0, -1, -2]
    >>> list(val for (val,) in values)
    [-2, -1, 0, 1, 2]
    >>> [what['lag'] for what in map(id2what, ids_sorted)]
    [-2, -1, 0, 1, 2]
    """
    whats = map(id2what, whatids)  # if this is bottleneck, allow to pass what themselves
    values = [whatvalues(what, keys) for what in whats]
    return tuple(zip(*[(whatid, value) for value, whatid in sorted(zip(values, whatids), key=itemgetter(0))]))


def call2what(depth=1, non_id_keys=None):
    """
    Returns a What instance with information about the locals of the caller at depth.

    With the current implementation this could not work on interpreters other than cpython.

    Parameters
    ----------
    depth : int, default 1
      The depth of the desired caller in the stack; 1 is the immediate caller of call2what

    non_id_keys : string iterable, default None
      What variables of the locals to consider as non-id when creating the What object.

    Examples
    --------
    >>> def f(x, y=3, z=5):
    ...     j = 33
    ...     print('With all locals', call2what().id())
    ...     print('Only keywords', call2what(non_id_keys=['x', 'j']).id())
    ...     return j, x + y + z
    >>> _ = f(2, z=12)
    With all locals f(j=33,x=2,y=3,z=12)
    Only keywords f(y=3,z=12)
    """
    # this would probably break on interpreters other than CPython
    frame, _, _, name, _, _ = inspect.stack()[depth]
    # we could probably be (too) clever here and get only the function parameter values
    # also we could add a parameter to automatically grab only keywords... for next iteration
    conf_dict = frame.f_locals
    return What(name, conf=conf_dict, non_id_keys=non_id_keys)


def _get_or_none(what, key):
    try:
        return what[key]
    except KeyError:
        return None


def _key2colname(key):
    return '_'.join(map(str, key)) if isinstance(key, (tuple, list)) else key


def whatid2columns(df, whatid_col, columns=None, prefix='', postfix='', inplace=True):
    """
    Extract values from whatami id strings into new columns in a pandas dataframe.

    Parameters
    ----------
    df : pandas DataFrame
      Should contain a column

    whatid_col : string
      The name of the column where the whatami id strings are stored.

    columns : list of strings or list of tuples of strings, default None
      The coordinates of the parameters to extract from the whatami ids (usually a string)
      If None, all the top-level parameters of the id string will be extracted.

    prefix : string
      The name of the new columns will be prefixed with this string (if None, no prefix)

    postfix : string
      The name of the new columns will be followed by this string (if None, no postfix)

    inplace : boolean, default True
      If False, make a copy of the dataframe and add the columns there; otherwise add the new columns to df.

    Returns
    -------
    df itself or a copy if inplace is False.
    """
    # parsing is slow, cache the Whats
    whats = {whatid: id2what(whatid) for whatid in df[whatid_col].unique()}

    if columns is None:
        columns = sorted(set(chain.from_iterable(what.keys() for what in whats.values())), key=_key2colname)

    prefix = '' if prefix is None else prefix
    postfix = '' if postfix is None else postfix

    if not inplace:
        df = df.copy()

    for column in columns:
        column_name = _key2colname(column)
        if isinstance(column, list):
            column = tuple(column)
        df[prefix + column_name + postfix] = df[whatid_col].apply(
            lambda whatid: _get_or_none(whats[whatid], column))

    return df

# --- Maintenance

OLD_WHATID_PARSER = None


def oldid2what(oldwhatid):
    """Parses an old-style whatami id into a What object.

    Examples
    --------
    >>> old_id = "out=acc_to_target#GoingTowards#im=True#positions=('x', 'y')#targets=(-0.1, -0.1)"
    >>> what = oldid2what(old_id)
    >>> print(what.id())
    GoingTowards(im=True,out='acc_to_target',positions=('x','y'),targets=(-0.1,-0.1))
    >>> old_id = "GoingTowards#im=True#positions=('x', 'y')#targets=('trg_x', 'trg_y')"
    >>> what = oldid2what(old_id)
    >>> print(what.id())
    GoingTowards(im=True,positions=('x','y'),targets=('trg_x','trg_y'))
    >>> old_id = "out=acc#GoingTowards#im=True#positions=('x', 'y')#targets=('trg_x', 'trg_y')#out='vel'"
    >>> what = oldid2what(old_id)
    Traceback (most recent call last):
    ...
    ValueError: whatid defines out ambiguously ("acc" and "vel")
    """
    global OLD_WHATID_PARSER
    if OLD_WHATID_PARSER is None:
        OLD_WHATID_PARSER = build_oldwhatami_parser()
    out = None
    if oldwhatid.startswith('out='):
        out, _, oldwhatid = oldwhatid.partition('#')
        out = out[4:]

    what = id2what(oldwhatid, parser=OLD_WHATID_PARSER)

    if out is not None:
        if 'out' in what.conf:
            raise ValueError('whatid defines out ambiguously ("%s" and "%s")' % (out, what['out']))
        what.conf['out'] = out

    return what


def id2whatami4(oldwhatid):
    """
    >>> print(id2whatami4("out=vel_to_target#GoingTowards#im=True#positions=('x', 'y')#targets=('trg_x', 'trg_y')"))
    GoingTowards(im=True,out='vel_to_target',positions=('x','y'),targets=('trg_x','trg_y'))
    >>> print(id2whatami4("velocity"))
    velocity
    """
    try:
        return oldid2what(oldwhatid).id()
    except NoMatch:
        return oldwhatid
